from typing import Dict, Any, Optional
import logging
import json
import base64
import struct
import hmac
import hashlib
import time

from ..core.constants import LOGGER_NAME
from ..core.constants import LOGIN
from ..core import Credentials, SessionCore, URLs, Config, PAClient
from ..core import check_session_config
from ..core.helpers import check_response
from ..core.helpers import camelcase_dict_to_snake
from ..core.helpers import ThrottlingClient

LOGGER = logging.getLogger(LOGGER_NAME)


# Dedicated Throttling client for login as it is much more restricted.
_LOGIN_THROTTLE = ThrottlingClient(max_requests=1, period_seconds=2)


async def login(
        credentials: Credentials,
        session: Optional[SessionCore] = None) -> SessionCore:
    """
    Authentify with Degiro API.
    `session` will be updated with required data for further connections.
    If no `session` is provided, create one.
    """
    url = URLs.LOGIN
    session = session or SessionCore()
    payload = {
        "username": credentials.username,
        "password": credentials.password,
        "isRedirectToMobile": False,
        "isPassCodeReset": '',
        "queryParams": {"reason": "session_expired"}
    }
    async with _LOGIN_THROTTLE as client:
        LOGGER.debug("login| url %s", url)
        response = await client.post(url, content=json.dumps(payload))
        LOGGER.debug("login| response %s", response.__dict__)

        response_load = response.json()

        if response_load['status'] == LOGIN.TOTP_NEEDED:
            # totp needed
            if (credentials.totp_secret is None and
                    credentials.one_time_password is None):
                raise AssertionError(
                    "Account has TOTP enabled, but no TOTP secret"
                    " nor one_time_password was provided.")
            elif credentials.totp_secret is not None:
                payload["oneTimePassword"] = _get_totp_token(
                    credentials.totp_secret)
            elif credentials.one_time_password is not None:
                payload["oneTimePassword"] = credentials.one_time_password

            url = URLs.LOGIN_TOTP
            LOGGER.debug("run totp login at %s", url)
            response = await client.post(
                url,
                content=json.dumps(payload),
                cookies=response.cookies)
            LOGGER.debug(response.__dict__)
            LOGGER.debug(response.json())

        check_response(response)
        session._cookies = response.cookies

        if SessionCore.JSESSIONID not in session._cookies:
            LOGGER.error("No JSESSIONID in response: %s", response)
            LOGGER.error("No JSESSIONID in response cookies: %s",
                         response.cookies)
            raise AssertionError("No JSESSIONID in response.")

        return session


async def get_config(session: SessionCore) -> SessionCore:
    """
    Populate session with configuration
    """
    _check_active_session(session)
    async with session as client:
        res = await client.get(URLs.CONFIG, cookies=session._cookies)

    check_response(res)
    config = Config(camelcase_dict_to_snake(res.json()['data']))

    session.config = config

    return session


async def get_client_info(session: SessionCore) -> SessionCore:
    """
    Get client information.
    """
    url = URLs.get_client_info_url(session)

    async with session as client:
        res = await client.get(
            url,
            params={'sessionId': session._cookies[session.JSESSIONID]},
            cookies=session._cookies)

    check_response(res)
    resp_data = res.json()['data']
    if 'id' in resp_data:
        resp_data['id'] = str(resp_data['id'])
    session.client = PAClient(camelcase_dict_to_snake(resp_data))
    return session


async def get_account_info(session: SessionCore) -> SessionCore:
    """

    """
    _check_active_session(session)
    url = URLs.get_account_info_url(session)
    async with session as client:
        res = await client.get(url,
                               cookies=session._cookies
                               )
    check_response(res)
    res_json = res.json()
    LOGGER.debug("get_account_info| res_json %s", res_json)
    return res_json


async def get_product_dictionary(session: SessionCore) -> Dict[str, Any]:
    """
    Get product dictionary information from server.

    This is needed to provide human-redeable product data for products:

    - Bonds, CFD Exchange places.
    - ETF fees types.
    - Countries.

    Returns
    -------

        Example JSON below, most rows were removed to keepsome structure
        readibility. Example from 2023.07.

        ``` JSON
        {'bondExchanges': [{'country': 905, 'exchange': 5001, 'id': 5001},
                           {'country': 846,
                            'exchange': 676,
                            'id': 676,
                            'postfix': 'NYSE'},
                           {'country': 957, 'exchange': 999, 'id': 999},
                           {'country': 846,
                            'exchange': 663,
                            'id': 663,
                            'postfix': 'NASDAQ'},
                           {'country': 915, 'exchange': 608, 'id': 608}],
         'bondIssuerTypes': [{'id': 0,
                              'name': 'all',
                              'translation': 'Koersen.step1.530'},
                             {'id': 2,
                              'name': 'company',
                              'translation': 'Koersen.step1.544'}],
         'bondSortColumns': [{'id': 'exchangeHiqAbbr'},
                             {'id': 'name'},
                             {'id': 'exchangeName'}],
         'cfdExchanges': [{'country': 906, 'exchange': 194, 'id': 194},
                          {'country': 975, 'exchange': 508, 'id': 508},
                          {'country': 915, 'exchange': 608, 'id': 608}],
         'cfdSortColumns': [{'id': 'exchangeHiqAbbr'},
                            {'id': 'name'},
                            {'id': 'exchangeName'}],
         'combinationExchanges': [{'id': 1, 'name': 'Euronext Amsterdam'},
                                  {'id': 2, 'name': 'Euronext Brussels'},
                                  {'id': 3, 'name': 'Euronext Paris'}],
         'countries': [{'id': 908,
                        'name': 'FI',
                        'region': 1,
                        'translation': 'list.country.908'},
                       {'id': 906,
                        'name': 'DE',
                        'region': 1,
                        'translation': 'list.country.906'}],
         'etfAggregateTypes': [{'id': 'quoteCurrency',
                                'name': 'quoteCurrency'},
                               {'id': 'totalExpenseRatioInterval',
                                'name': 'totalExpenseRatioInterval'},
                               {'id': 'administrator',
                                'name': 'administrator'},
                               {'id': 'etpType', 'name': 'etpType'},
                               {'id': 'exchange', 'name': 'exchange'},
                               {'id': 'region', 'name': 'region'},
                               {'id': 'assetAllocation',
                                'name': 'assetAllocation'},
                               {'id': 'issuer', 'name': 'issuer'},
                               {'id': 'benchmark', 'name': 'benchmark'}],
         'etfFeeTypes': [{'id': 1,
                          'name': 'Trackers',
                          'translation': 'contract.type.131'},
                         {'id': 2,
                          'name': 'Free',
                          'translation': 'translation.label.388'}],
         'etfSortColumns': [{'id': 'quoteCurrency'},
                            {'id': 'administrator'},
                            {'id': 'etpType'},
                            {'id': 'exchangeHiqAbbr'},
                            {'id': 'name'},
                            {'id': 'exchangeName'},
                            {'id': 'region'},
                            {'id': 'assetAllocation'},
                            {'id': 'benchmark'}],
         'eurexCountries': [{'exchanges': [{'city': 'Helsinki',
                                            'code': 'HE',
                                            'country': 'FI',
                                            'hiqAbbr': 'HSE',
                                            'id': 1001,
                                            'micCode': 'XHEL',
                                            'name': 'OMX Helsinki'}],
                             'id': 908,
                             'name': 'FI',
                             'underlyingExchangeIds': [1001]},
                            {'exchanges': [{'city': 'Frankfurt',
                                            'code': 'XETR',
                                            'country': 'DE',
                                            'hiqAbbr': 'XET',
                                            'id': 194,
                                            'micCode': 'XETR',
                                            'name': 'Xetra'}],
                             'id': 906,
                             'name': 'DE',
                             'underlyingExchangeIds': [194]}],
         'exchanges': [{'city': 'Frankfurt',
                        'code': 'CATS',
                        'country': 'DE',
                        'hiqAbbr': 'CAT',
                        'id': 190,
                        'micCode': 'CATS',
                        'name': 'CATS'},
                       {'city': 'Amsterdam',
                        'country': 'XX',
                        'hiqAbbr': 'FSM',
                        'id': 10000,
                        'name': 'FundShare Matching'},
                       {'city': 'Copenhagen',
                        'code': 'CO',
                        'country': 'DK',
                        'hiqAbbr': 'OMK',
                        'id': 508,
                        'micCode': 'XCSE',
                        'name': 'OMX Copenhagen'}],
         'futureExchanges': [{'id': 1, 'name': 'Euronext Liffe'},
                             {'eurexCountries': [908,
                                                 915,
                                                 906],
                              'id': 2,
                              'name': 'Eurex'},
                             {'id': 9, 'name': 'COMEX'}],
         'futureSortColumns': [{'id': 'exchangeHiqAbbr'},
                               {'id': 'name'},
                               {'id': 'exchangeName'},
                               {'id': 'expirationDate'}],
         'indices': [{'id': 106002, 'name': 'SDAX'},
                     {'id': 106001, 'name': 'MDAX'},
                     {'id': 121003, 'name': 'SMIM', 'productId': 11875105},
                     {'id': 114003, 'name': 'ISEQ Overall'},
                     {'id': 121002, 'name': 'SLI', 'productId': 11875104}],
         'investmentFundAggregateTypes': [{'id': 'issuer', 'name': 'issuer'}],
         'investmentFundFeeTypes': [{'id': 0,
                                     'name': 'Rebate-free',
                                     'translation': 'translation.label.387'},
                                    {'id': 3,
                                     'name': 'Rebate-free FREE',
                                     'translation': 'translation.label.386'}],
         'investmentFundSortColumns': [{'id': 'exchangeHiqAbbr'},
                                       {'id': 'name'},
                                       {'id': 'exchangeName'}],
         'leveragedAggregateTypes': [{'id': 'shortLong', 'name': 'shortLong'},
                                     {'id': 'underlyingProductId',
                                      'name': 'underlyingProductId'},
                                     {'id': 'exchange', 'name': 'exchange'},
                                     {
                                        'id': 'underlying',
                                        'name': 'underlying'
                                     },
                                     {'id': 'issuer', 'name': 'issuer'},
                                     {'id': 'expirationDateRange',
                                      'name': 'expirationDateRange'}],
         'leveragedSortColumns': [{'id': 'leverage'},
                                  {'id': 'financingLevel'},
                                  {'id': 'stoploss'},
                                  {'id': 'exchangeHiqAbbr'},
                                  {'id': 'name'},
                                  {'id': 'exchangeName'},
                                  {'id': 'expirationDate'}],
         'lookupSortColumns': [{'id': 'exchangeHiqAbbr'},
                               {'id': 'name'},
                               {'id': 'exchangeName'},
                               {'id': 'productType'}],
         'optionAggregateTypes': [{'id': 'month', 'name': 'month'},
                                  {'id': 'derivativeExchange',
                                   'name': 'derivativeExchange'},
                                  {'id': 'year', 'name': 'year'},
                                  {'id': 'strike', 'name': 'strike'}],
         'optionExchanges': [{'exchangeId': 365, 'id': 11, 'name': 'NYMEX'},
                             {'exchangeId': 1006,
                              'id': 1,
                              'name': 'Euronext Amsterdam',
                              'underlyingExchangeIds': [200, 212, 3001]},
                             {'eurexCountries': [908,
                                                 982,
                                                 906],
                              'exchangeId': 185,
                              'id': 5,
                              'name': 'Eurex',
                              'underlyingExchangeIds': [1001,
                                                        320,
                                                        608,
                                                        194]},
                             {'exchangeId': 359, 'id': 9, 'name': 'CME'}],
         'optionSortColumns': [{'id': 'strike'}, {'id': 'expirationDate'}],
         'productTypes': [{'briefTranslation': 'contract.type.brief.13',
                           'contractType': 'FND',
                           'id': 13,
                           'name': 'fund',
                           'translation': 'contract.type.13'},
                          {'briefTranslation': 'contract.type.brief.8',
                           'contractType': 'OPT',
                           'id': 8,
                           'name': 'option',
                           'translation': 'contract.type.8'}],
         'regions': [{'id': 1,
                      'name': 'Europe',
                      'translation': 'translation.label.117'},
                     {'id': 2,
                      'name': 'America',
                      'translation': 'translation.label.118'},
                     {'id': 3,
                      'name': 'Other',
                      'translation': 'translation.label.121'}],
         'stockCountries': [{'country': 908,
                             'id': 908,
                             'indices': [112002, 16, 112001, 120005]},
                            {'country': 849, 'id': 849, 'indices': [124001]},
                            {'country': 905, 'id': 905, 'indices': [123001]},
                            {'country': 906,
                             'id': 906,
                             'indices': [6, 106002, 106003, 106001]}],
         'stockSortColumns': [{'id': 'exchangeHiqAbbr'},
                              {'id': 'name'},
                              {'id': 'exchangeName'}]}
        ```
    """
    check_session_config(session)
    url = URLs.get_product_dictionary_url(session)
    params = dict(
        intAccount=session.client.int_account,
        sessionId=session.config.session_id
    )
    async with session as client:
        response = await client.get(url,
                                    cookies=session._cookies,
                                    params=params
                                    )
    check_response(response)
    LOGGER.debug("webapi.get_product_dictionary response| %s", response.json())
    return response.json()


###########
# Helpers #
###########

def _check_active_session(session: SessionCore):
    """
    Check that session id has been populated. Raise AssertionError if not.
    """
    if SessionCore.JSESSIONID not in session._cookies:
        raise AssertionError("No JSESSIONID in session.cookies")


def _get_totp_token(secret_key: str) -> str:
    "Get one-time-password from secret key"
    key = base64.b32decode(secret_key)
    message = struct.pack(">Q", int(time.time()) // 30)
    message_hash = hmac.new(key, message, hashlib.sha1).digest()
    o = message_hash[19] & 15
    message_hash = (struct.unpack(">I",
                    message_hash[o:o + 4])[0] & 0x7fffffff) % 1000000
    return message_hash


__all__ = [
    login.__name__,
    get_account_info.__name__,
    get_config.__name__
]



from typing import Union, List, Dict, Any
from typing import Set
import sys
from typing import Optional
import functools
import logging
import pprint
import copy

from jsonloader import JSONclass

from ..core import LOGGER_NAME
from ..core import SessionCore
from ..core import Config
from ..core import PAClient
from ..core import Credentials
from ..core import camelcase_dict_to_snake
from ..core import BadCredentialsError
from ..core import lru_cache_timed
from ..core import PRODUCT
from .. import webapi


LOGGER = logging.getLogger(LOGGER_NAME)


@JSONclass(annotations=True, annotations_type=True)
class Region:
    id: int
    name: str


@JSONclass(annotations=True, annotations_type=True)
class Country:
    id: str
    name: str  # ISO 3166-1 alpha-2 country codename
    region: 'Region'


@JSONclass(annotations=True, annotations_type=True)
class Exchange:
    id: str
    name: str
    city: Optional[str] = None
    code: Optional[str] = None
    country_name: str  # renamed from 'country' as it is country name
    hiq_abbr: str
    mic_code: Optional[str] = None


@JSONclass(annotations=True, annotations_type=True)
class Index:
    # Example of JSON
    # {
    #   "data": {
    #     "4824940": {
    #       "id": "4824940",
    #       "name": "CAC 40",
    #       "isin": "FR0003500008",
    #       "symbol": "CAC INDEX",
    #       "contractSize": 1,
    #       "productType": "INDEX",
    #       "productTypeId": 180,
    #       "tradable": false,
    #       "category": "H",
    #       "currency": "EUR",
    #       "active": true,
    #       "exchangeId": "710",
    #       "onlyEodPrices": false,
    #       "orderTimeTypes": [],
    #       "buyOrderTypes": [],
    #       "sellOrderTypes": [],
    #       "productBitTypes": [],
    #       "closePrice": 7384.91,
    #       "closePriceDate": "2023-07-20",
    #       "feedQuality": "R",
    #       "orderBookDepth": 0,
    #       "vwdIdentifierType": "issueid",
    #       "vwdId": "360015511",
    #       "qualitySwitchable": false,
    #       "qualitySwitchFree": false,
    #       "vwdModuleId": 1
    #     }
    #   }
    # }
    @JSONclass(annotations=True)
    class Info:
        id: str
        isin: str
        name: str
        symbol: str
        product_type: PRODUCT.TYPE
        product_type_id: PRODUCT.TYPEID
        currency: str
        exchange_id: str
        exchange: Exchange
        close_price: Optional[float] = None
        close_price_date: Optional[str] = None
        vwd_id: Optional[str] = None
        vwd_identifier_type: Optional[str] = None

    id: str
    name: str
    product_id: Optional[str] = None
    info: Optional[Info] = None

    async def get_info(self, session: 'Session') -> Optional[Info]:
        """
        Populate and return Index.Info. Provides more information about Index,
        such as isin, symbol, exchange ...

        If no Info is available for this Index, returns None.

        In the event you need to instantiate all or most indices, you should
        use :class:`~degiroasync.api.Session.populate_indices_info` that
        will reduce web API calls.

        See :class:`~degiroasync.api.Index.Info` for details.
        """
        if self.info is None and self.product_id is None:
            return None
        elif self.info is None and self.product_id is not None:
            ind_id = self.product_id
            resp = await webapi.get_products_info(
                    session,
                    products_ids=[ind_id],
                    )
            # Info needs to be populated.
            LOGGER.debug("Index.get_info| %s", resp)
            self.info = self._dict2info(session, resp['data'][ind_id])
        return self.info

    @classmethod
    def _dict2info(
            cls: type,
            session: 'Session',
            web_response: Dict[str, Any]) -> Info:
        info_dict = camelcase_dict_to_snake(web_response)
        info_dict['product_type_id'] = PRODUCT.TYPEID(
                info_dict['product_type_id'])
        info_dict['product_type'] = PRODUCT.TYPE(
                info_dict['product_type'])
        info_dict['exchange'] = session.dictionary.exchange_by(
                id=info_dict['exchange_id'])
        return cls.Info(info_dict)


class ExchangeDictionary:
    """
    Usage:

    .. code-block:: python

        exchangedict = await ExchangeDictionary(session)
        exchangedict.exchange_by(hiqAbbr='EPA')

    """

    exchanges: List[Exchange]
    countries: List[Country]
    regions: List[Region]
    indices: List[Index]

    _exchanges: Dict[str, Any]

    async def __new__(cls, session: SessionCore):
        self = super().__new__(cls)

        product_dictionary = await webapi.get_product_dictionary(session)
        LOGGER.debug("api.ExchangeDictionary| %s",
                     pprint.pformat(product_dictionary))
        self._regions = {p['id']: Region(p)
                         for p in product_dictionary['regions']}

        self._countries_id = {}
        self._countries_name = {}
        for country in product_dictionary['countries']:
            # Replace region dict by object.
            region = self._regions[country['region']]
            country['region'] = region
            country['id'] = str(country['id'])
            # Register country
            country_inst = Country(country)
            self._countries_id[country['id']] = country_inst
            self._countries_name[country['name']] = country_inst

        self._exchanges = {}
        for exchange in product_dictionary['exchanges']:
            # Some APIs call return id in int, others in str. Fix here as all
            # str.
            exchange['id'] = str(exchange['id'])
            # Replace region dict by object.
            exchange['countryName'] = exchange['country']
            del exchange['country']

            # Register country
            self._exchanges[exchange['id']] = Exchange(
                    camelcase_dict_to_snake(exchange))

        self._indices = {}
        self._indices_name = {}
        for index in product_dictionary['indices']:
            index = camelcase_dict_to_snake(index)
            index['id'] = str(index['id'])
            if 'product_id' in index:
                index['product_id'] = str(index['product_id'])
            self._indices[index['id']] = Index(index)
            self._indices_name[index['name'].lower()] = Index(index)

        return self

    @property
    def exchanges(self):
        return [v for v in self._exchanges.values()]

    @property
    def countries(self):
        return [c for c in self._countries_name.values()]

    @property
    def regions(self):
        return [r for r in self._regions.values()]

    @property
    def indices(self):
        return [i for i in self._indices.values()]

    @functools.lru_cache(32)
    def index_by(
            self,
            name: Optional[str] = None,
            id: Optional[int] = None,
            ) -> Index:
        if name is not None and id is not None:
            raise AssertionError("Exactly one of (name, id) must be not None.")
        if name is not None:
            return self._indices_name[name.lower()]
        if id is not None:
            return self._indices[id]

    @functools.lru_cache(32)
    def exchange_by(
            self,
            *,
            name: Optional[str] = None,
            id: Optional[int] = None,
            hiq_abbr: Optional[str] = None,
            mic_code: Optional[str] = None) -> Exchange:
        """
        Get Exchange by *either* name, hiqAbbr (e.g. EPA),
        micCode (e.g. XPAR).
        """
        if sum(attr is not None for attr in (name,
                                             id,
                                             hiq_abbr,
                                             mic_code)) != 1:
            raise AssertionError(
                "Exactly one of (name, id, hiqAbbr, micCode) "
                "must be not None.")
        if id is not None:
            return self._exchanges[id]
        for exc in self._exchanges.values():
            if name is not None and exc.name == name:
                return exc
            elif hiq_abbr is not None and exc.hiq_abbr == hiq_abbr:
                return exc
            elif mic_code is not None and exc.mic_code == mic_code:
                return exc
        raise KeyError("No exchange found with search attributes: {}",
                       (name, id, hiq_abbr, mic_code))

    def country_by(
            self,
            *,
            name: Union[str, None] = None,
            id: Union[int, None] = None) -> Country:
        if sum(attr is not None for attr in (name, id)) != 1:
            raise AssertionError(
                "Exactly one of (name, id) must be not None.")
        if name is not None:
            if name not in self._countries_name:
                raise KeyError(
                        "{} not found. List of available countries: {}".format(
                            name,
                            [c for c in self._countries_name]
                            ))
            return self._countries_name[name]
        if id is not None:
            return self._countries_id[id]

    async def populate_indices_info(self, session: 'Session'):
        """
        Populate info attribute for all indices.
        """
        # Get ids
        ids = [
                i.product_id
                for i in self._indices.values()
                if i.product_id
                ]
        # Query data
        resp = await webapi.get_products_info(session, ids)
        # Populate info parameter for all indices
        for index in self._indices.values():
            if index.product_id:
                index.info = Index._dict2info(
                        session,
                        resp['data'][index.product_id]
                        )


class Session(SessionCore):
    config: Config
    client: PAClient
    dictionary: ExchangeDictionary

    def __init__(self,
                 session_core: SessionCore,
                 dictionary: ExchangeDictionary):
        """
        Use `api.login` to get a populated Session instance.

        Build a `Session` instance from a `SessionCore` and an
        ExchangeDictionary.
        """
        LOGGER.debug("Session.__init__: session_core %s", session_core)
        if session_core.config is None:
            raise AssertionError("session_core.config not set.")
        if session_core.client is None:
            raise AssertionError("session_core.client not set.")
        self.__dict__.update(copy.copy(session_core.__dict__))
        self.config = session_core.config
        self.client = session_core.client
        self.dictionary = dictionary

    @property
    def exchange_dictionary(self):
        print(
            'Session.exchange_dictionary is deprecated. Please use '
            'Session.dictionary instead.', file=sys.stderr)
        return self.dictionary


async def get_dictionary(session: Session) -> ExchangeDictionary:
    """
    Return a new ExchangeDictionary from Session.
    """
    return await ExchangeDictionary(session)


#: Singleton to store hash of failed attempts credentials.
_LOGIN_FAILURE_HASH: Set[str] = set()

#: Minimum time before new login attempt is allowed with bad credentials (s).
#: Default = 3 hours
_LOGIN_FAILURE_MINTIME = 60 * 60 * 3


@lru_cache_timed(maxsize=32, seconds=_LOGIN_FAILURE_MINTIME)
def _should_fail(credentials: Credentials) -> bool:
    if credentials in _LOGIN_FAILURE_HASH:
        # We can discard failure as it will be remembered by this function
        # cache.
        _LOGIN_FAILURE_HASH.discard(credentials)
        return True
    else:
        return False


async def login(
        credentials: Credentials,
        session: Union[Session, None] = None,
        /,
        safeguard_incorrect_credentials: bool = True
        ) -> Session:
    """
    Authentify with Degiro API and populate basic information that'll be needed
    for further calls.

    `session` will be updated with required data for further connections.

    Roughly equivalent to:

    .. code-block:: python

        session = await degiroasync.webapi.login(credentials)
        await webapi.get_config(session)
        await webapi.get_client_info(session)

    If no `session` is provided, create one.

    Parameters
    ----------

    credentials:
        Credentials with which we attempt login.

    session:
        If provided, populate with API config. If not a new Session will be
        created.

    safeguard_incorrect_credentials:
        Defaults to True. If `True`, store credentials hash in case of failure
        and do not allow new login attempts with same hash. This allows to
        avoid blocking account in case of inaccurate credentials.

    Raises
    ------

    BadCredentialsError:
        If login with those credentials returned a bad credentials response.

    ResponseError:
        Other response errors.

    """
    if safeguard_incorrect_credentials:
        if credentials in _LOGIN_FAILURE_HASH:
            LOGGER.error("Bad credentials, abort login.")
            raise BadCredentialsError(
                    "Abort. Login previously failed with these credentials."
                    " Provide `safeguard_incorrect_credentials=False` "
                    "if you want to override this behavior.")

    try:
        session_core = await webapi.login(credentials)
    except BadCredentialsError:
        LOGGER.error("Bad credentials, abort login.")
        _LOGIN_FAILURE_HASH.add(credentials)
        raise

    await webapi.get_config(session_core)
    await webapi.get_client_info(session_core)
    dictionary = await get_dictionary(session_core)
    if session is not None:
        session.__dict__.update(copy.deepcopy(session_core))
        session.dictionary = dictionary
    else:
        return Session(session_core, dictionary)
    return session


def check_session_dictionary(session: Session):
    """
    Check that session exchange dictionary has been populated.
    Raise AssertionError if not.
    """
    if session.dictionary is None:
        raise AssertionError(
            "session.dictionary is not set. "
            "Use api.login to build your session or call ")



import enum

try:
    # 2022.04 available starting Python 3.11. We're within a try/except
    # with an alternative, ignore type.
    from enum import StrEnum  # type: ignore
except ImportError:
    # Exists only starting Python 3.11
    # Reimplement what we need from it here.
    class StrEnum(str, enum.Enum):  # type: ignore
        def __str__(self):
            return str.__str__(self)


# Used to get same Logger instance across submodules.
LOGGER_NAME = 'degiroasync'


class _EnumBase(enum.Enum):
    """
    Base class for Enums that should behave as str or other types.

    >>> class StrEnum(str, _EnumBase):
    ...     pass
    ...
    >>> class Foo(StrEnum):
    ...     A = 'foo'
    ...
    >>> str(Foo.A)
    'foo'

    """

    def __repr__(self):
        return f"'{self.__str__()}'"

    def __str__(self):
        return self.value.__str__()


class IntEnum(int, _EnumBase):
    """
    Base class for Enums that should behave as str.

    Refer to example below for difference in behavior with other Enum based
    solutions.

    >>> import enum
    >>> # Basic enum.Enum example
    >>> class Foo(enum.Enum):
    ...     A = 1
    ...
    >>> Foo.A
    <Foo.A: 1>
    >>> str(Foo.A)
    'Foo.A'
    >>> repr(Foo.A)
    "<Foo.A: 1>"
    >>> type(Foo.A)
    <enum 'Foo'>
    >>> # Inheriting from str, enum.Enum
    >>> class Foo(int, enum.Enum):
    ...     A = 1
    ...
    >>> Foo.A
    <Foo.A: 1>
    >>> str(Foo.A)
    'Foo.A'
    >>> repr(Foo.A)
    "<Foo.A: 1>"
    >>> type(Foo.A)
    <enum 'Foo'>

    >>> # An StrEnum child can be used as a drop-in replacement for an int
    >>> class Foo(IntEnum):
    ...     A = 1
    ...
    >>> Foo.A
    1
    >>> str(Foo.A)
    '1'
    >>> repr(Foo.A)
    "'1'"
    >>> type(Foo.A)
    <enum 'Foo'>
    """


class LOGIN(IntEnum):
    """
    Login-related constants.

    At this date, only one has been noticed and documented.

    TOTP_NEEDED
        This flag is set when one-time-password is required to log in the
        account.
    """
    TOTP_NEEDED = 6
    BAD_CREDENTIALS = 3


class ORDER:
    """
    Constants for orders in web API.
    """
    class ACTION(StrEnum):
        """
        BUY
            Use when placing a _BUY_ order on a product.

        SELL:
            Use when placing a _SELL_ order on a product.

        Check `ORDER.TYPE` for types of orders.

        For use with `checkOrder` HTTP endpoint - or `check_order` degiroasync
        API calls.
        """
        BUY = 'BUY'
        SELL = 'SELL'

    class TYPE(IntEnum):
        """
        LIMITED
            This is the "Limit" field value of "Order type" in the web trader.

        MARKET
            This is the "Market" field value of "Order type" in the web trader.

        STOP_LOSS
            This is the "Stop Loss" field value of "Order type" in the web
            trader.

        STOP_LIMITED
            This is the "Stop Limited" field value of "Order type" in the web
            trader.

        AMOUNT
            This is the "Amount" field value of "Order type" in the web trader.
            It's used for investment funds.
        """
        LIMITED = 0
        STOP_LIMITED = 1
        MARKET_ORDER = 2
        STOP_LOSS = 3
        AMOUNT = 4

    class TIME(IntEnum):
        """
        DAY
            This is the "Day" option in the web trader.
            Order placed for one day only: it expires at the end of the day it
            was created.

        PERMANENT
            This is the "GTC" option in the web trader.
            This order sticks until it has been executed. Be careful that
            occasionally Degiro cancels Permanent orders due to various events.
        """
        DAY = 1
        PERMANENT = 3

    class STATUS(StrEnum):
        """
        CONFIRMED
            The order was accepted by the platform.

        REJECTED
            The order was *not* accepted by the platform.

        PENDING
            The order is 'PENDING' on the web trader.
        """
        CONFIRMED = 'CONFIRMED'
        REJECTED = 'REJECTED'
        PENDING = 'PENDING'


class TRANSACTION:
    class TYPEID(IntEnum):
        pass

    class COUNTERPARTY(StrEnum):
        """
        Counterparty in web trader transactions details.

        So far, 3 counterparts were documented, see below.
        It is not completely clear when one counterparty is used rather than
        the other. Check DEGIRO tool documentation for more details should
        this be needed.

        """
        MARKET = 'MK'
        DEGIRO = 'DG'
        GROUP = 'GR'


class PRODUCT:
    class TYPEID(IntEnum):
        """
        STOCK
            Stocks, or shares products.

        BONDS
            Bonds products.

        FUTURES
            Futures products.

        OPTIONS
            Options products

        FUNDS
            Investment funds products.

        LEVERAGE_PRODUCTS
            Leverage products.

        ETFS
            ETFs products.

        CFDS
            CFDs products.

        WARRANTS
            Warrants products.

        CURRENCY
            Currencies. This is used by endpoint for account balances for
            instance.

        """
        STOCK = 1
        INDEX = 180
        BONDS = 2
        FUTURES = 7
        OPTIONS = 8
        FUNDS = 13
        LEVERAGE_PRODUCTS = 14
        ETFS = 131
        CFDS = 535
        WARRANTS = 536
        CURRENCY = 311

    class TYPE(StrEnum):
        STOCK = 'STOCK'
        INDEX = 'INDEX'
        FUND = 'FUND'
        CASH = 'CASH'
        CFD = 'CFD'
        CURRENCY = 'CURRENCY'


class SORT(StrEnum):
    """
    Used by some endpoints to provide sorted results.

    ASC
        Results will be sorted in ascending order.

    DESC
        Results will be sorted in descending order.
    """
    ASC = "asc"
    DESC = "desc"


class PRICE:
    class RESOLUTION(StrEnum):
        # Resolution
        PT1M = 'PT1M'  # Higher resolution, 1 minute
        PT1D = 'P1D'   # 1 tic per day

    class PERIOD(StrEnum):
        # Periods
        # P stands for Prior?
        P1DAY = 'P1D'
        P1WEEK = 'P1W'
        P1MONTH = 'P1M'
        P3MONTH = 'P3M'
        P6MONTH = 'P6M'
        P1YEAR = 'P1Y'
        P3YEAR = 'P3Y'
        P5YEAR = 'P5Y'
        P50YEAR = 'P50Y'

    class TYPE(StrEnum):
        PRICE = 'price'
        OHLC = 'ohlc'


class POSITION:
    class TYPE(StrEnum):
        PRODUCT = 'PRODUCT'


# HTTP Requests global timeout, in seconds. DEGIRO endpoints are sometimes
# a bit slow to answer, maybe due to endpoint throttling: this is necessary
# to limit number of call failure.
TIMEOUT: float = 30



import logging

from .core import Credentials
from .core import Config
from .core import ResponseError
from .core import LOGGER_NAME
from . import webapi
from . import api

LOGGER = logging.getLogger(LOGGER_NAME)
LOGGER.setLevel(logging.DEBUG)

__all__ = [
        # Choice between strings or import errors at this level.
        # An object missing __name__ will fail at module import, typos
        # in object names are flagged by editor's tools and easier to catch
        # than typos in a string.
        obj.__name__ for obj in (  # type: ignore
            Credentials,
            Config,
            api,
            webapi,
            ResponseError,
        )
] + ['LOGGER_NAME']



from typing import Optional
import logging
import os
import asyncio

import degiroasync.api
from degiroasync.api import Session
from degiroasync.core import ResponseError
from degiroasync.core import Credentials

LOGGER = logging.getLogger(degiroasync.core.LOGGER_NAME)

def _get_credentials():
    """
    Helper to get credentials for integration tests
    """
    username = os.environ.get('DEGIRO_USERNAME')
    password = os.environ.get('DEGIRO_PASSWORD')
    assert username is not None, (
        'DEGIRO_USERNAME environment variable not defined.')
    assert password is not None, (
        'DEGIRO_PASSWORD environment variable not defined.')
    totp_secret = os.environ.get('DEGIRO_TOTP_SECRET')

    return Credentials(username, password, totp_secret)


class _IntegrationLogin:
    """
    Internal helper, can be inherited to make login for integration tests
    easier.
    """

    _lock = asyncio.Lock()
    session: Optional[Session] = None
    _login_attempted: bool = False

    @classmethod
    async def _login(cls):
        async with cls._lock:
            if cls.session is None and not cls._login_attempted:
                LOGGER.debug("_IntegrationLogin: attempt login.")
                cls._login_attempted = True
                credentials = _get_credentials()
                cls.session = await degiroasync.api.login(credentials)
        if cls.session is None:
            raise ResponseError("No session available. Maybe Bad Credentials?")
        return cls.session



import logging
from typing import Union, Dict, Any
import datetime

from .product import get_trading_update
from ..core import SessionCore
from ..core import URLs
from ..core import constants
from ..core import ORDER
from ..core import join_url
from ..core import check_session_client
from ..core import check_session_config
from ..core.helpers import check_response
from ..core.helpers import dict_from_attr_list


LOGGER = logging.getLogger(constants.LOGGER_NAME)


async def confirm_order(
        session: SessionCore,
        *,
        confirmation_id: str,
        product_id: str,
        buy_sell: str,
        time_type: ORDER.TIME,
        order_type: ORDER.TYPE,
        size: int,
        price: Union[float, None] = None,
):
    """
    Close the placing order process. This is used after `check_order`
    to confirm placing an order. It must be called with the same paramaters
    and the confirmation_id obtained with `check_order`.

    buy_sell:
        One of ('BUY', 'SELL').

    price:
        Price in currency of the product.

    product_id:
        Product on which to place the order.

    order_type:
        Specify if we want a STOP_LIMITED, MARKET_ORDER, STOP_LOSS,
        STOP_LIMITED.

    time_type:
        Specify if we want a DAY or PERMANENT order.

    Check `degiroasync.core.ORDER` for details.

    Example data with request:

    .. code-block:: python
        {
            "buySell": "BUY",
            "orderType": 0,
            "price": 10,
            "productId": "96008",
            "size": 1,
            "timeType": 1
        }

    Example JSON response:

    .. code-block:: python

        {
            'data': {
            'confirmationId': 'a8e49a7e-4d79-4f21-961b-988dc0806d09',
            'freeSpaceNew': 55395,
            'transactionFee': 0.5,
            'showExAnteReportLink': True
            }
        }

    """
    # This call will not have integration tests to prevent misplaced orders
    # All changes must be manually verified to work as intended.
    _order_calls_check(session,
                       product_id=product_id,
                       buy_sell=buy_sell,
                       time_type=time_type,
                       order_type=order_type,
                       size=size,
                       price=price)

    jsession_id = session._cookies[session.JSESSIONID]

    url = join_url(
        URLs.get_confirm_order_url(session),
        confirmation_id
    )
    url += f';{jsession_id}'

    params = dict(
        intAccount=session.client.int_account,
        sessionId=jsession_id
    )
    data = dict(
        buySell=buy_sell,
        orderType=order_type,
        price=price,
        productId=product_id,
        timeType=time_type
    )

    async with session as client:
        response = await client.post(
            url,
            params=params,
            json=data,
        )
    check_response(response)
    resp_json = response.json()
    LOGGER.debug("check_order| %s", resp_json)
    return resp_json


def _order_calls_check(
        session: SessionCore,
        *,
        product_id: str,
        buy_sell: str,
        time_type: ORDER.TIME,
        order_type: ORDER.TYPE,
        size: int,
        price: Union[float, None] = None,
):
    check_session_config(session)
    check_session_client(session)

    # We're a bit more strict than usual in type checking here: it's really
    # *not* the place where we want to give room and flexibility to users
    # in exchange of safety.
    if not isinstance(order_type, ORDER.TYPE):
        raise TypeError("order_type must be an ORDER.TYPE")

    if not isinstance(time_type, ORDER.TIME):
        raise TypeError("order_type must be an ORDER.TIME")

    if order_type != ORDER.TYPE.MARKET_ORDER and price is None:
        raise AssertionError("price must not be None for orders that are not "
                             "MARKET_ORDER.")


async def check_order(
        session: SessionCore,
        *,
        product_id: str,
        buy_sell: str,
        time_type: ORDER.TIME,
        order_type: ORDER.TYPE,
        size: int,
        price: Union[float, None] = None,
) -> Dict[str, Any]:
    """
    Start the placing order process. This is used to retrieve order
    cost for instance and is mandatory to actually place an order with
    `confirm_order`.

    buy_sell:
        One of ('BUY', 'SELL').

    price:
        Price in currency of the product.

    product_id:
        Product on which to place the order.

    order_type:
        Specify if we want a STOP_LIMITED, MARKET_ORDER, STOP_LOSS,
        STOP_LIMITED.

    time_type:
        Specify if we want a DAY or PERMANENT order.

    Check `degiroasync.core.ORDER` for details.

    Example data with request:
    ```
    {
        "buySell": "BUY",
        "orderType": 0,
        "price": 10,
        "productId": "96008",
        "size": 1,
        "timeType": 1
    }
    ```

    Example JSON response:
    ```
    {
        "data": {
            "confirmationId": "c64332f0-a800-4f04-bc29-12a2b7efa20b",
            "freeSpaceNew": 2282.0,
            "showExAnteReportLink": true,
            "transactionFee": 0.5
        }

    }
    ```
    """
    if buy_sell not in ("BUY", "SELL"):
        raise AssertionError("buy_sell not 'BUY' or 'SELL'")
    LOGGER.debug("webapi.check_order| %s", str(buy_sell))
    _order_calls_check(session,
                       product_id=product_id,
                       buy_sell=buy_sell,
                       time_type=time_type,
                       order_type=order_type,
                       size=size,
                       price=price)

    jsessionid = session._cookies[session.JSESSIONID]

    url = URLs.get_check_order_url(session)
    params = dict(
        intAccount=session.client.int_account,
        sessionId=jsessionid
    )
    data = dict(
        buySell=buy_sell,
        orderType=order_type,
        price=price,
        productId=product_id,
        size=size,
        timeType=time_type
    )
    LOGGER.debug("check_order data| %s", data)
    async with session as client:
        response = await client.post(
            url,
            params=params,
            json=data,
            headers={
                'content-type': 'application/json;charset=UTF-8'
            },
            cookies=session._cookies
        )
    check_response(response)
    resp_json = response.json()
    LOGGER.debug("check_order| %s", resp_json)
    return response.json()


async def get_orders(session: SessionCore) -> Dict[str, Any]:
    """
    Get current and historical orders.

    Example response:

    {'orders': [
        {
        'id': '52401b7d-00ed-4aed-979d-a2476e331b26',
        'date': '2022-05-12T15:08:33',
        'productId': 255931,
        'product': 'Veolia Environnement',
        'contractType': 1,
        'contractSize': 1.0,
        'currency': 'EUR',
        'buysell': 'B',
        'size': 100.0,
        'quantity': 100.0,
        'price': 24.7,
        'stopPrice': 0.0,
        'totalOrderValue': 2470.0,
        'orderTypeId': 0,
        'orderTimeTypeId': 1,
        'orderType': 'LIMIT',
        'orderTimeType': 'DAY',
        'isModifiable': True,
        'isDeletable': True
        }]
    }

    """
    # TODO: assess if historicalOrders is relevant here
    # or if it should be removed and we should only rely on get_orders_history
    orders = await get_trading_update(
        session,
        params={
            'orders': 0,
            }
        # 'historicalOrders': 0}
    )
    LOGGER.debug("webapi.get_orders| orders %s", orders)
    orders = {
            'orders': [
                dict_from_attr_list(order['value'])
                for order in orders['orders']['value']
                if order['name'] == 'order'
                ]
            }
    LOGGER.debug("webapi.get_orders| orders rebuilt %s", orders)
    return orders


ORDER_DATE_FORMAT = '%d/%m/%Y'


async def get_orders_history(
        session: SessionCore,
        from_date: str,
        to_date: str,
) -> Dict[str, Any]:
    """
    Get historical orders for session.

    from_date:
        Date in format DD/MM/YYYY. Raise ValueError if incorrect format.
    to_date:
        Date in format DD/MM/YYYY. Raise ValueError if incorrect format.

    Example return in json():
    ```
    {
        'data': {
            [
                {
                    'created': '2021-11-05T11:56:33+01:00',
                    'orderId': 'asdfjkhrewqrkljh',
                    'productId': 12341234,
                    'size': 100,
                    'price': 390,
                    'buysell': 'B',
                    'orderTypeId': 1,
                    'stopPrice': 0,
                    'currentTradedSize': 0,
                    'totalTradedSize': 0,
                    'type': 'CREATE',
                    'status': 'CONFIRMED',
                    'last': '2021-11-6T12:33:23+01:00',
                    'isActive': True
                }
            ]
        }
    }
    ```
    """
    check_session_config(session)
    check_session_client(session)
    # Check date format, datetime will raise an exception
    datetime.datetime.strptime(from_date, ORDER_DATE_FORMAT)
    datetime.datetime.strptime(to_date, ORDER_DATE_FORMAT)

    jsessionid = session._cookies[session.JSESSIONID]

    url = URLs.get_orders_history_url(session)
    params = dict(
        fromDate=from_date,
        toDate=to_date,
        intAccount=session.client.int_account,
        sessionId=jsessionid
    )

    async with session as client:
        response = await client.get(
            url,
            params=params,
            headers={
                'content-type': 'application/json'
            },
            cookies=session._cookies
        )
    check_response(response)
    resp_json = response.json()
    LOGGER.debug("get_orders_history| %s", resp_json)
    return response.json()


async def get_transactions(
        session: SessionCore,
        from_date: str,
        to_date: str,
) -> Dict[str, Any]:
    """
    Get transactions for session.

    from_date:
        Date in format DD/MM/YYYY. Raise ValueError if incorrect format.
    to_date:
        Date in format DD/MM/YYYY. Raise ValueError if incorrect format.

    Example return in json():
    ```
    {
        'data': {
            [
                {
                    'id': 123123,
                    'productId': 12341234,
                    'date': '2021-11-05T11:56:33+01:00',
                    'buysell': 'B',
                    'price': 39,
                    'quantity': 100,
                    'total': -3900,
                    'orderTypeId': 1,
                    'counterParty': 1,
                    'transfered': False,
                    'fxRate': 0,
                    'totalInBaseCurrency': -3900,
                    'feeInBaseCurrency': -12,
                    'totalPlusFeeInBaseCurrency': -3912,
                    'transactionTypeId': -3912,
                    'tradingVenue': -3912,
                }
            ]
        }
    }
    ```
    """
    check_session_config(session)
    check_session_client(session)
    # Check date format, datetime will raise an exception
    datetime.datetime.strptime(from_date, ORDER_DATE_FORMAT)
    datetime.datetime.strptime(to_date, ORDER_DATE_FORMAT)

    jsessionid = session._cookies[session.JSESSIONID]

    url = URLs.get_transactions_url(session)
    params = dict(
        fromDate=from_date,
        toDate=to_date,
        intAccount=session.client.int_account,
        sessionId=jsessionid,
        groupTransactionsByOrder=False
    )

    LOGGER.debug('get_transactions params| %s', params)
    async with session as client:
        response = await client.get(
            url,
            params=params,
            headers={
                'content-type': 'application/json'
            },
            cookies=session._cookies
        )
    check_response(response)
    resp_json = response.json()
    LOGGER.debug("get_transactions response| %s", resp_json)
    return resp_json



"""
High-level API for DEGIRO platform.
===================================

This module provides an interface that is higher level than the `webapi` module
(or the direct HTTP API).
It provides various data structures with guaranteed fields, and functions
to build or act on those data structures.

Users should only import directly from `degiroasync.api`, its submodules
should be considered implementation details.

Data Structures
+++++++++++++++

Details on data structures are available for each data structure.

Core structures to be aware off when dealing with `degiroasync` are:

    - *Session*
      It is necessary with almost every call in this module. Session contains
      session identifier, as well as basic mapping information between
      identifiers and objects.
    - *Product*
      A product per DEGIRO API. This can be a Shares, ETF or any other product
      available on the platform.
    - *Order*
      An order per DEGIRO API. As in the web trader, an order can be active
      or not, and different order types exist (BUY, SELL, LIMITED ...).
    - *Transaction*
      A transaction per DEGIRO API. This reflects cash impacts on the account.
      There can be zero to many transactions per Order. As of 2022.03, DEGIRO
      API does not seem to provide relationship between transactions and
      orders.

Other data structures are available, but more situational. They can be found
in the module members documentation.

Functions
+++++++++

Several functions reflecting the API calls are implemented in this module,
An overview of some of the essential ones is listed below.
Details can be found with each function documentation:

    - *login*
      Builds a `Session` instance, this supports login with or without 2FA.
    - *get_portfolio* and *get_portfolio_total*
      Provide data on products currently held on the account.
    - *search_product*
      Find and instantiate DEGIRO products from various search options (text,
      ISIN, symbol, exchange ...).
    - *get_price_series*
      Get price data on a Product.
    - *get_orders* and *check_order*
      Get orders on the account, start placing an order. Note: `webapi` module
      must currently be used to confirm an order.

"""
from .orders import check_order
from .orders import get_orders
from .orders import get_transactions
from .orders import Order
from .orders import Transaction
from .session import Session
from .session import Index
from .session import login
from .session import get_dictionary
from .session import ExchangeDictionary
from .session import Exchange, Region, Country
from .product import get_portfolio
from .product import get_portfolio_total
from .product import get_price_series
from .product import get_price_data
from .product import search_product
from .product import Stock
from .product import Currency
from .product import ProductBase
from .product import ProductFactory  # Don't expose in __all__
from .product import ProductGeneric
from .product import PriceSeries
from .product import TotalPortfolio
from ..core import Credentials
from ..core.constants import PRODUCT
from ..core.constants import PRICE
from ..core.constants import ORDER
from ..core.constants import POSITION


__all__ = [
        # Choice between strings or import errors at this level.
        # An object missing __name__ will fail at module import, typos
        # in object names are flagged by editor's tools and easier to catch
        # than typos in a string.
        obj.__name__ for obj in (  # type: ignore
            # Session
            login,
            Credentials,
            Session,
            get_dictionary,
            ExchangeDictionary,
            Exchange,
            Region,
            Country,
            Index,

            # Product
            search_product,
            get_portfolio,
            get_price_data,
            get_price_series,

            # PriceData,
            PriceSeries,
            Stock,
            Currency,
            ProductBase,
            ProductGeneric,

            get_portfolio,
            get_portfolio_total,
            TotalPortfolio,

            # Order
            check_order,
            get_orders,
            get_transactions,
            Order,
            Transaction,

            # Constants
            # PRODUCT,
            # ORDER,
            # PRICE,
            # POSITION
        )
        ]



import unittest
import logging
import os
import pprint
import asyncio
import datetime
from typing import Optional
import unittest.mock
from unittest.mock import MagicMock

import httpx

import degiroasync
import degiroasync.webapi
import degiroasync.core
import degiroasync.core.helpers
from degiroasync.core import Credentials
from degiroasync.core import SessionCore
from degiroasync.core.constants import PRODUCT
from degiroasync.core.constants import PRICE
from degiroasync.core.constants import ORDER
from degiroasync.core.constants import LOGIN
from degiroasync.core import ResponseError
from degiroasync.core import BadCredentialsError
from degiroasync.core.helpers import ThrottlingClient
from degiroasync.webapi import get_config
from degiroasync.webapi import get_products_info
from degiroasync.webapi import get_company_profile
from degiroasync.webapi import get_news_by_company


LOGGER = logging.getLogger(degiroasync.core.LOGGER_NAME)

RUN_INTEGRATION_TESTS = 0
try:
    _env_var = os.environ.get('DEGIROASYNC_INTEGRATION')
    RUN_INTEGRATION_TESTS = int(_env_var)
except (ValueError, TypeError):
    LOGGER.info('degiroasync integration tests will *not* run.')

RUN_BAD_CREDENTIALS = 0
try:
    _env_var = os.environ.get('DEGIROASYNC_INTEGRATION_BAD_CREDENTIALS')
    RUN_BAD_CREDENTIALS = int(_env_var)
except (ValueError, TypeError):
    LOGGER.info('degiroasync Bad Credentials integration test will *not* run.')


class TestDegiroAsyncWebAPI(unittest.IsolatedAsyncioTestCase):
    @unittest.mock.patch('degiroasync.core.helpers.ThrottlingClient.post')
    async def test_login_bad_credentials(
            self,
            post_m
            ):
        """
        Verify that BadCredentialsError is raised when endpoint returns a bad
        credentials error.
        """
        post_m.return_value = MagicMock()
        response = post_m.return_value
        response.json = MagicMock(
                return_value={
                    'status': LOGIN.BAD_CREDENTIALS,
                    'content': 'badCredentials'})
        response.status_code = httpx.codes.BAD_REQUEST

        credentials = Credentials(
            username='dummyaccount123456',
            password='dummydummy'
                )
        with self.assertRaises(BadCredentialsError):
            await degiroasync.webapi.login(credentials)


if RUN_INTEGRATION_TESTS:
    LOGGER.info('degiroasync.webapi integration tests will run.')
    from tests.integration_login import _IntegrationLogin

    class TestDegiroAsyncWebAPIIntegration(
            _IntegrationLogin,
            unittest.IsolatedAsyncioTestCase):

        async def test_login(self):
            session = await _IntegrationLogin._login()
            self.assertTrue('JSESSIONID' in session.cookies,
                            "No JSESSIONID found.")

        if RUN_BAD_CREDENTIALS:
            async def test_login_bad_credentials(self):
                credentials = Credentials(
                    username='dummyaccount123456',
                    password='dummydummy'
                        )
                with self.assertRaises(BadCredentialsError):
                    await degiroasync.webapi.login(credentials)

        async def test_config(self):
            session = await _IntegrationLogin._login()
            await get_config(session)
            LOGGER.debug('test_config| %s', session.config)
            self.assertTrue(
                    session.config.pa_url is not None,
                    "paUrl not defined.")
            self.assertTrue(
                    session.config.product_search_url is not None,
                    "productSearchUrl not defined.")
            self.assertTrue(
                    session.config.trading_url is not None,
                    "tradingUrl not defined.")

        async def test_portfolio(self):
            session = await _IntegrationLogin._login()

            resp_json = await degiroasync.webapi.get_portfolio(session)
            LOGGER.debug("test_portfolio| %s", resp_json)
            self.assertTrue('portfolio' in resp_json)
            self.assertTrue('value' in resp_json['portfolio'])

        async def test_portfolio_total(self):
            session = await _IntegrationLogin._login()

            resp_json = await degiroasync.webapi.get_portfolio_total(session)
            LOGGER.debug("test_portfolio_total| %s", resp_json)
            self.assertTrue('totalPortfolio' in resp_json)
            self.assertTrue('value' in resp_json['totalPortfolio'])

        async def test_get_products_info(self):
            session = await _IntegrationLogin._login()

            resp_json = await degiroasync.webapi.get_portfolio(session)
            portfolio = resp_json['portfolio']
            product_ids = filter(lambda x: x is not None,
                                 (product.get('id')
                                  for product in portfolio['value']))
            response = await get_products_info(session,
                                               [p for p in product_ids])
            self.assertIsInstance(response, dict)

            response = await degiroasync.webapi.get_products_info(session,
                                                                  ["72906"])
            self.assertIsInstance(response, dict)
            LOGGER.debug('webapi.test_get_products_info| %s',
                         pprint.pformat(response))

        async def test_get_company_profile(self):
            session = await _IntegrationLogin._login()

            isin = "NL0000235190"
            resp_json = await get_company_profile(session, isin)
            self.assertTrue('data' in resp_json, resp_json)
            self.assertTrue('businessSummary' in resp_json['data'], resp_json)
            LOGGER.debug('webapi.test_get_company_profile| %s',
                         pprint.pformat(resp_json))

        async def test_get_news_by_company(self):
            session = await _IntegrationLogin._login()

            isin = "NL0000235190"
            resp_json = await get_news_by_company(session, isin)
            self.assertTrue('data' in resp_json, resp_json)
            self.assertTrue('items' in resp_json['data'], resp_json)

        async def test_get_price_series(self):
            """
            Simply check that we don't have an error and data is not empty.
            """
            session = await _IntegrationLogin._login()

            vwdId = '360114899'

            resp_json = await degiroasync.webapi.get_price_series(
                    session,
                    vwdId=vwdId,
                    vwdIdentifierType='issueid')
            LOGGER.debug('get_price_series response: %s', resp_json)
            LOGGER.debug(resp_json)
            self.assertIn('series', resp_json)
            self.assertIn('data', resp_json['series'][0])

        async def test_get_price_series_month(self):
            session = await _IntegrationLogin._login()

            vwdId = '360114899'

            resp_json = await degiroasync.webapi.get_price_series(
                    session,
                    vwdId=vwdId,
                    period=PRICE.PERIOD.P1MONTH,
                    resolution=PRICE.RESOLUTION.PT1M,
                    vwdIdentifierType='issueid')
            LOGGER.debug(resp_json)
            self.assertIn('resolution', resp_json)
            self.assertEqual(resp_json['resolution'], PRICE.RESOLUTION.PT1M)
            self.assertIn('series', resp_json)
            self.assertIn('data', resp_json['series'][0])

        async def test_get_price_series_ohlc(self):
            session = await _IntegrationLogin._login()

            vwdId = '360114899'

            resp_json = await degiroasync.webapi.get_price_series(
                    session,
                    vwdId=vwdId,
                    period=PRICE.PERIOD.P1MONTH,
                    resolution=PRICE.RESOLUTION.PT1M,
                    vwdIdentifierType='issueid',
                    data_type=PRICE.TYPE.OHLC,
                    )
            LOGGER.debug(resp_json)
            self.assertIn('resolution', resp_json)
            self.assertEqual(resp_json['resolution'], PRICE.RESOLUTION.PT1M)
            self.assertIn('series', resp_json)
            self.assertIn('data', resp_json['series'][0])
            self.assertEqual(
                    len(resp_json['series'][0]['data'][0]), 5,
                    "We should have 5 entries per row (index + O H L C)")

        async def test_get_price_series_month_pt1d(self):
            session = await _IntegrationLogin._login()

            vwdId = '360114899'

            resp_json = await degiroasync.webapi.get_price_series(
                    session,
                    vwdId=vwdId,
                    period=PRICE.PERIOD.P1MONTH,
                    resolution=PRICE.RESOLUTION.PT1D,
                    vwdIdentifierType='issueid')
            LOGGER.debug('get_price_series_month_pt1d| response: %s',
                         resp_json)
            LOGGER.debug(resp_json)
            self.assertIn('resolution', resp_json)
            self.assertEqual(resp_json['resolution'], PRICE.RESOLUTION.PT1D)
            self.assertIn('series', resp_json)
            self.assertIn('data', resp_json['series'][0])

        async def test_search_product(self):
            session = await _IntegrationLogin._login()

            search = "AIRBUS"
            resp_json = await degiroasync.webapi.search_product(
                    session,
                    search)
            self.assertIn('products', resp_json, resp_json)
            self.assertGreaterEqual(len(resp_json['products']), 1)
            self.assertIn('id', resp_json['products'][0], resp_json)
            self.assertIn('isin', resp_json['products'][0], resp_json)
            self.assertIn('name', resp_json['products'][0], resp_json)

        #async def test_search_product_exchange(self):
        #    raise NotImplementedError()
        #    session = await _IntegrationLogin._login()

        #    search = "AIRBUS"
        #    resp_json = await degiroasync.webapi.search_product(
        #            session,
        #            search)
        #    self.assertIn('products', resp_json, resp_json)
        #    self.assertGreaterEqual(len(resp_json['products']), 1)
        #    self.assertIn('id', resp_json['products'][0], resp_json)
        #    self.assertIn('isin', resp_json['products'][0], resp_json)
        #    self.assertIn('name', resp_json['products'][0], resp_json)

        async def test_search_product_by_index(self):
            session = await _IntegrationLogin._login()
            resp_json = await degiroasync.webapi.get_product_dictionary(
                    session,
                    )

            index = session.dictionary.index_by(name='CAC 40')
            LOGGER.debug("test_search_product_by_index| Index: %s", index)
            resp_json = await degiroasync.webapi.search_product(
                    session,
                    index_id=index.id,
                    limit=100
                    )
            self.assertIn('products', resp_json, resp_json)
            # We should have 40 products in CAC 40
            self.assertGreaterEqual(len(resp_json['products']), 40)

        async def test_product_dictionary(self):
            session = await _IntegrationLogin._login()

            resp_json = await degiroasync.webapi.get_product_dictionary(
                    session
                    )

            self.assertIn('exchanges', resp_json)
            self.assertIn('countries', resp_json)
            self.assertIn('regions', resp_json)

            # Not used by degiroasync.api at the time this test was written.
            self.assertIn('bondExchanges', resp_json)
            self.assertIn('cfdExchanges', resp_json)
            self.assertIn('combinationExchanges', resp_json)
            self.assertIn('etfAggregateTypes', resp_json)
            self.assertIn('etfFeeTypes', resp_json)
            self.assertIn('eurexCountries', resp_json)

    class TestDegiroWebAPIOrdersIntegration(
            _IntegrationLogin,
            unittest.IsolatedAsyncioTestCase):
        """
        Set Orders will *not* be tested: this would imply being charged every
        time tests are executed.
        """
        async def test_get_orders(self):
            session = await _IntegrationLogin._login()

            resp_json = await degiroasync.webapi.get_orders(session)
            LOGGER.debug("test_get_orders| %s", pprint.pformat(resp_json))
            self.assertIn('orders', resp_json)
            self.assertIsInstance(resp_json['orders'], list)

        async def test_check_order(self):
            session = await _IntegrationLogin._login()

            # Leverage api.search_product to get a specific product_id
            # as example for integration testing of check_order.
            # This is introducing a dependency on api module, but is easier
            # to manage. Future improvement opportunit for this test: implement
            # required query and filter here using only webapi.
            from degiroasync import api
            session.dictionary = await api.get_dictionary(
                    session)
            products = await api.search_product(
                    session,
                    by_symbol="AIR",
                    by_exchange="EPA",
                    product_type_id=PRODUCT.TYPEID.STOCK
                    )
            self.assertEqual(len(products), 1,
                             "We should only have one product here")
            product = products[0]
            # Reminder: typeId = 1 is STOCK
            LOGGER.debug("test_check_order| product: %s . %s . %s . %s",
                         product.info.symbol,
                         product.info.name,
                         product.info.product_type_id,
                         product.base.id)

            # This will *not* place the order: it would have to be confirmed
            # with `confirm_order` call.
            resp_json = await degiroasync.webapi.check_order(
                    session,
                    product_id=product.base.id,
                    buy_sell=ORDER.ACTION.BUY,
                    time_type=ORDER.TIME.DAY,
                    order_type=ORDER.TYPE.LIMITED,
                    size=1,
                    price=50
                    )
            LOGGER.debug("test_check_order| %s", pprint.pformat(resp_json))
            self.assertIn('data', resp_json)
            self.assertIn('confirmationId', resp_json['data'])
            self.assertIn('freeSpaceNew', resp_json['data'])
            self.assertIn('transactionFee', resp_json['data'])

            resp_json = await degiroasync.webapi.check_order(
                    session,
                    product_id=product.base.id,
                    buy_sell=ORDER.ACTION.SELL,
                    time_type=ORDER.TIME.DAY,
                    order_type=ORDER.TYPE.LIMITED,
                    size=1,
                    price=50
                    )
            LOGGER.debug("test_check_order| %s", pprint.pformat(resp_json))
            self.assertIn('data', resp_json)
            self.assertIn('confirmationId', resp_json['data'])
            self.assertIn('freeSpaceNew', resp_json['data'])
            self.assertTrue(
                    'transactionFee' in resp_json['data']
                    or 'transactionOppositeFee' in resp_json['data'],
                    resp_json['data'])

        async def test_get_account_info(self):
            session = await _IntegrationLogin._login()
            resp_json = await degiroasync.webapi.get_account_info(session)
            LOGGER.debug("test_get_account_info| response: %s", resp_json)

            self.assertIn('data', resp_json)
            self.assertIn('clientId', resp_json['data'])
            self.assertIn('baseCurrency', resp_json['data'])
            # Not sure what more to test here. To be extended when this call
            # usage has been identified.

        async def test_get_orders_history(self):
            session = await _IntegrationLogin._login()
            to_date = datetime.datetime.today()
            from_date = datetime.datetime.today() - datetime.timedelta(days=7)
            date_format = degiroasync.webapi.orders.ORDER_DATE_FORMAT
            resp_json = await degiroasync.webapi.get_orders_history(
                    session,
                    from_date=from_date.strftime(date_format),
                    to_date=to_date.strftime(date_format)
                    )
            LOGGER.debug("test_get_orders_history| response: %s", resp_json)

            self.assertIn('data', resp_json)
            data = resp_json['data']
            for order in data:
                self.assertIn('created', order)
                self.assertIn('productId', order)
                self.assertIn('size', order)
                self.assertIn('price', order)
                self.assertIn('buysell', order)
                self.assertIn(order['buysell'], ('B', 'S'))
                self.assertIn('orderTypeId', order)
                self.assertIn('orderTimeTypeId', order)
                self.assertIn('type', order)
                self.assertIn('status', order)
                self.assertIn('last', order)
                self.assertIn('isActive', order)
                self.assertIn('currentTradedSize', order)
                self.assertIn('totalTradedSize', order)

        async def test_get_orders_history_date_check(self):
            session = await _IntegrationLogin._login()
            with self.assertRaises(ValueError):
                await degiroasync.webapi.get_orders_history(
                    session,
                    from_date='garbage',
                    to_date='garbage')

        async def test_get_transactions(self):
            session = await _IntegrationLogin._login()
            to_date = datetime.datetime.today()
            from_date = datetime.datetime.today() - datetime.timedelta(days=7)
            date_format = degiroasync.webapi.orders.ORDER_DATE_FORMAT
            resp_json = await degiroasync.webapi.get_transactions(
                    session,
                    from_date=from_date.strftime(date_format),
                    to_date=to_date.strftime(date_format)
                    )
            LOGGER.debug("test_get_orders_history| response: %s", resp_json)

            self.assertIn('data', resp_json)
            data = resp_json['data']
            for trans in data:
                self.assertIn('id', trans)
                self.assertIn('productId', trans)
                self.assertIn('quantity', trans)
                self.assertIn('price', trans)
                self.assertIn('fxRate', trans)
                self.assertIn('nettFxRate', trans)
                self.assertIn('transfered', trans)
                self.assertIn('buysell', trans)
                self.assertIn(trans['buysell'], ('B', 'S'))
                self.assertIn('transactionTypeId', trans)

        async def test_get_transactions_date_check(self):
            session = await _IntegrationLogin._login()
            with self.assertRaises(ValueError):
                await degiroasync.webapi.get_orders_history(
                    session,
                    from_date='garbage',
                    to_date='garbage')


if __name__ == '__main__':
    handler = logging.StreamHandler()
    handler.setLevel(logging.DEBUG)
    formatter = logging.Formatter(
            "%(asctime)s-%(name)s-%(levelname)s-%(message)",
            "%Y%m%d"
            )
    if handler not in LOGGER.handlers:
        LOGGER.addHandler(handler)
    LOGGER.setLevel(logging.DEBUG)
    unittest.main()



"""
Contains core data structures and helpers for degiroasync.

Key structures that are reused regularly through degiroasync API the below:
    - Credentials
    - SessionCore
    - ORDER
    - PRODUCT
    - TRANSACTION
    - POSITION
    - ResponseError

Check out their respective documentation for more details.
"""
from .helpers import join_url
from .helpers import lru_cache_timed
from .helpers import camelcase_to_snake
from .helpers import camelcase_dict_to_snake
from .helpers import set_params
from .constants import ORDER
from .constants import PRODUCT
from .constants import TRANSACTION
from .constants import PRICE
from .constants import POSITION
from .constants import LOGGER_NAME
from .core import check_session_client
from .core import check_session_config
from .core import Credentials
from .core import Config
from .core import SessionCore
from .core import URLs
from .core import PAClient
from .exceptions import ResponseError
from .exceptions import BadCredentialsError
from .exceptions import ContextError


# Fail at module import if an attribute is missing __name__: ignore mypy here.
__all__ = [obj.__name__ for obj in (  # type: ignore
    join_url,
    lru_cache_timed,
    camelcase_to_snake,
    camelcase_dict_to_snake,
    set_params,
    PRODUCT,
    ORDER,
    TRANSACTION,
    POSITION,
    PRICE,
    Credentials,
    Config,
    check_session_client,
    check_session_config,
    ResponseError,
    BadCredentialsError,
    ContextError,
    SessionCore,
    PAClient,
    URLs,
    )] + [
        LOGGER_NAME
    ]



from __future__ import annotations
from typing import List, Dict, Any, Iterable, Callable, Coroutine
from typing import Optional
import logging
import sys
import functools
import asyncio
import time


import asyncstdlib.functools as afunctools
import httpx

from .constants import LOGGER_NAME
from .constants import LOGIN
from .exceptions import ResponseError
from .exceptions import BadCredentialsError


LOGGER = logging.getLogger(LOGGER_NAME)
# Logs helpers
FORMAT_DEFAULT = '%(asctime)s-%(name)s-%(levelname)s- %(message)s'
STREAMHANDLER_DEFAULT = logging.StreamHandler(stream=sys.stdout)


class CoroCache:
    def __init__(self, coro: Coroutine):
        """
        Cache to allow awaiting coroutines several times.

        >>> async def f():
        ...     return 1
        ...
        >>> c = CoroCache(f())
        >>> await c
        1
        >>> await c
        1

        """
        self.coro = coro
        self.res: Optional[Any] = None
        self.wait: bool = True

    def __await__(self) -> Any:
        if self.wait:
            self.res = self.coro.__await__()
            self.wait = False
        return self.res


def lru_cache_timed(
        func: Optional[Callable] = None,
        /,
        maxsize: int = 128,
        typed: bool = False,
        seconds: Optional[float] = None,
):
    """
    Time-sensitive LRU cache that works with async functions.

    >>> @lru_cache_timed(seconds=120)
    ... async def foo():
    ...     asyncio.sleep(1)
    ...     return 1
    ...
    >>> import time
    >>> start = time.time()
    >>> await foo()
    1
    >>> time.time() - start > 1
    True
    >>> start = time.time()
    >>> await foo()
    1
    >>> time.time() - start < .1
    True
    """
    first_start: List[float] = []
    if func is not None:
        if asyncio.iscoroutinefunction(func):
            @afunctools.lru_cache(maxsize=maxsize, typed=typed)
            async def _in(time_key, *args, **kwargs):
                return await func(*args, **kwargs)

            @functools.wraps(func)
            async def _out(*args, **kwargs):
                if not len(first_start):
                    first_start.append(time.time())
                if seconds is None:
                    time_key = 1
                else:
                    time_key = (time.time() - first_start[0]) // seconds

                return await _in(time_key, *args, **kwargs)

            return _out

        else:
            @functools.lru_cache(maxsize=maxsize, typed=typed)
            def _in(time_key, *args, **kwargs):
                return func(*args, **kwargs)

            @functools.wraps(func)
            def _out(*args, **kwargs):
                if not len(first_start):
                    first_start.append(time.time())
                if seconds is None:
                    time_key = 1
                else:
                    time_key = (time.time() - first_start[0]) // seconds

                return _in(time_key, *args, **kwargs)

            return _out
    else:
        return lambda func: lru_cache_timed(func,
                                            maxsize=maxsize,
                                            typed=typed,
                                            seconds=seconds)


def camelcase_to_snake(text: str) -> str:
    """
    Convert a camelCase text to snake_case.

    This helper replace any capitalized character by its lowered version
    preceded by '_'.

    This helper does not check that input text is camel case or not.

    >>> s = 'iAmCamelCase'
    >>> camelcase_to_snake(s)
    'i_am_camel_case'
    >>> # Ignore fully uppercase text.
    >>> camelcase_to_snake("ALL_CAPS")
    'ALL_CAPS'
    """
    if text.isupper():
        return text
    return ''.join(map(lambda c: '_' + c.lower() if c.isupper() else c, text))


def camelcase_dict_to_snake(
        dict_in: Dict[str, Any],
        /,
        recursive: bool = True) -> Dict[str, Any]:
    """
    Convert keys of dictionary with `str` keys from camelCase to snake_case.

    This does not care for structures with depth: values dictionaries will not
    be updated to snake_case.

    >>> d = {'fooBar': 2, 'camelCase': {'camelCase': 1}}
    >>> camelcase_dict_to_snake(d)
    {'foo_bar': 2, 'camel_case': {'camel_case': 1}}
    >>> camelcase_dict_to_snake(d, recursive=False)
    {'foo_bar': 2, 'camel_case': {'camelCase': 1}}
    """
    if not recursive:
        return {
            camelcase_to_snake(k): v
            for k, v in dict_in.items()
        }
    else:
        return {
            camelcase_to_snake(k):
                camelcase_dict_to_snake(v, recursive=True)
                if isinstance(v, dict) else v
            for k, v in dict_in.items()
        }


def join_url(*sections):
    """
    Helper to build urls, with slightly different behavior from
    urllib.parse.urljoin, see example below.

    >>> join_url('https://foo.bar', '/rest/of/url')
    'https://foo.bar/rest/of/url'
    >>> join_url('https://foo.bar/product', '/rest/of/url')
    'https://foo.bar/product/rest/of/url'
    >>> join_url('https://foo.bar/product/', '/rest/of/url')
    'https://foo.bar/product/rest/of/url'
    >>> # We need this helper because of below
    ... # urllib.parse.urljoin behavior that is not suitable
    ... # for our purpose.
    >>> import urllib.parse
    >>> urllib.parse.urljoin('https://foo.bar/product', '/rest//of/url')
    'https://foo.bar/rest/of/url'

    """
    return '/'.join(s.strip('/') for s in sections)


def check_response(response: httpx.Response):
    """
    Raises
    ------

    BadCredentialsError:
        If return code == 400 and web API status returns bad credentials code.
    ResponseError:
        If return code != 200
    """
    if response.status_code == httpx.codes.BAD_REQUEST:
        resp_json = response.json()
        if (
                'status' in resp_json
                and resp_json['status'] == LOGIN.BAD_CREDENTIALS
                ):
            raise BadCredentialsError(
                    f"Bad Credentials. Reponse content: "
                    f"{str(response.content)}")

    if response.status_code not in (httpx.codes.OK, httpx.codes.CREATED):
        raise ResponseError(
            f"Error on call: url {response.url}"
            f" | code {response.status_code} | "
            f"content {str(response.content)}")


def dict_from_attr_list(
        attributes_list: Iterable[Dict[str, Any]],
        ignore_error=False) -> Dict[str, Any]:
    """
    Helper to build a dict from an attribute list as returned by the web API.

    attributes_list is expected to be an iterable of dict with 'isAdded',
    'name', 'value'. See below for an example.
    ```
            [
                 {'isAdded': True,
                  'name': 'id',
                  'value': '8614787'},
                 {'isAdded': True,
                  'name': 'positionType',
                  'value': 'PRODUCT'},
                 {'isAdded': True,
                  'name': 'size',
                  'value': 100},
                 {'isAdded': True,
                  'name': 'price',
                  'value': 73.0},
                 {'isAdded': True,
                  'name': 'value',
                  'value': 7300.0},
                 {'isAdded': True,
                  'name': 'accruedInterest'},
                 {'isAdded': True,
                  'name': 'plBase',
                  'value': {'EUR': -6717.901595272}},
                 {'isAdded': True,
                  'name': 'todayPlBase',
                  'value': {'EUR': -7300.0}},
                 {'isAdded': True,
                  'name': 'portfolioValueCorrection',
                  'value': 0},
                 {'isAdded': True,
                  'name': 'breakEvenPrice',
                  'value': 68.01},
                 {'isAdded': True,
                  'name': 'averageFxRate',
                  'value': 1},
                 {'isAdded': True,
                  'name': 'realizedProductPl',
                  'value': 97.098302728},
                 {'isAdded': True,
                  'name': 'realizedFxPl',
                  'value': 0},
                 {'isAdded': True,
                  'name': 'todayRealizedProductPl',
                  'value': 0.0},
                 {'isAdded': True,
                  'name': 'todayRealizedFxPl',
                  'value': 0}
            ]
    ```

    """
    dict_out = {}
    for attr in attributes_list:
        if 'name' not in attr or 'value' not in attr:
            message = (f"A provided params does not have a 'name' or 'value': "
                       f"{attr}")
            LOGGER.info(message)
            if not ignore_error:
                raise ValueError(message)
            else:
                continue
        name, value = attr['name'], attr['value']
        if not hasattr(name, int.__hash__.__name__):
            raise ValueError(f"{name} in {attr} is not hashable:"
                             " it can't be used as dict key.")
        dict_out[name] = value
        # TODO: above should never raise an error as long as attr['name'] is
        # hashable. May
        # try:
        #    dict_out[attr['name']] = attr['value']
        # except Exception as exc:
        #    LOGGER.error('dict_from_attr| %s: %s', type(exc), exc.args)
        #    if not ignore_error:
        #        raise exc
    return dict_out


def set_params(
        obj: Any,
        attributes_list: Iterable[Dict[str, str]],
        ignore_error=False) -> Any:
    """
    This is a helper to translate Degiro-format params to a Python object.

    Set `params` on `obj`.

    `params` is expected to be an iterable of dict with 'isAdded', 'name',
    'value'. See below for an example.

    ```
            [
                 {'isAdded': True,
                  'name': 'id',
                  'value': '8614787'},
                 {'isAdded': True,
                  'name': 'positionType',
                  'value': 'PRODUCT'},
                 {'isAdded': True,
                  'name': 'size',
                  'value': 100},
                 {'isAdded': True,
                  'name': 'price',
                  'value': 73.0},
                 {'isAdded': True,
                  'name': 'value',
                  'value': 7300.0},
                 {'isAdded': True,
                  'name': 'accruedInterest'},
                 {'isAdded': True,
                  'name': 'plBase',
                  'value': {'EUR': -6717.901595272}},
                 {'isAdded': True,
                  'name': 'todayPlBase',
                  'value': {'EUR': -7300.0}},
                 {'isAdded': True,
                  'name': 'portfolioValueCorrection',
                  'value': 0},
                 {'isAdded': True,
                  'name': 'breakEvenPrice',
                  'value': 68.01},
                 {'isAdded': True,
                  'name': 'averageFxRate',
                  'value': 1},
                 {'isAdded': True,
                  'name': 'realizedProductPl',
                  'value': 97.098302728},
                 {'isAdded': True,
                  'name': 'realizedFxPl',
                  'value': 0},
                 {'isAdded': True,
                  'name': 'todayRealizedProductPl',
                  'value': 0.0},
                 {'isAdded': True,
                  'name': 'todayRealizedFxPl',
                  'value': 0}
            ]
    ```

    """
    for k, v in dict_from_attr_list(
            attributes_list, ignore_error=ignore_error).items():
        setattr(obj, k, v)

    return obj


def setattrs(obj: Any, **attributes) -> Any:
    "Set all attributes on obj."
    for k, v in attributes.items():
        setattr(obj, k, v)
    return obj


def check_keys(data: dict, keys: Iterable[str]):
    """
    Helper to check keys are in data. Raise KeyError if that is not the case
    """
    for attr in keys:
        if attr not in data:
            name = 'data'
            if hasattr(data, '__name__'):
                name = data.__name__
            exc = KeyError(
                "{attr} not found in {name}. Data content: {data}.".format(
                    attr=attr,
                    name=name,
                    data=data
                ))
            raise exc


class ThrottlingClient:
    def __init__(
            self,
            *,
            max_requests=10,
            period_seconds=1,
            **kwargs
            ):
        """
        Wraps httpx.AsyncClient and throttle requests.


        Parameters
        ----------

        max_requests
            Maximum number of requests per `period_seconds` before throttling.
            If <= 0, no limit.

        period_seconds
            Period on which to count requests.

        kwargs
            Additional keywords are passed to httpx.AsyncClient

        """
        self._max_requests = max_requests
        self._period_s = period_seconds
        self._requests_times = []
        self._async_client: Optional[httpx.AsyncClient] = None
        self._client_open: Optional[httpx.AsyncClient] = None
        self._count_open = 0
        self._kwargs = kwargs

    def _throttle(method):
        "Decorator to wrap method to add throttling capabilities."
        @functools.wraps(method)
        async def wrapper(self, *args, **kwargs):
            LOGGER.debug(
                    "ThrottlingClient._throttle: "
                    "len(self._requests_times) %s",
                    len(self._requests_times))
            while (
                    self._max_requests > 0 and
                    len(self._requests_times) >= self._max_requests
                    ):
                # Clean register
                to_del = 0
                for ind, timestamp in enumerate(self._requests_times):
                    if time.time() - timestamp > self._period_s:
                        to_del += 1
                while to_del > 0:
                    self._requests_times.pop(0)
                    to_del -= 1
                # requests_times might be empty now.
                if len(self._requests_times) > 0:
                    time_delta = (self._requests_times[0] + self._period_s
                                  - time.time())
                    if time_delta > 0:
                        LOGGER.debug(
                                "Throttle %s call for %.2f",
                                method,
                                time_delta)
                        await asyncio.sleep(time_delta)
            if self._max_requests > 0:
                self._requests_times.append(time.time())
            return await method(self, *args, **kwargs)
        return wrapper

    async def __aenter__(self):
        if self._async_client is None:
            self._async_client = httpx.AsyncClient(**self._kwargs)
        if self._client_open is None:
            self._client_open = await self._async_client.__aenter__()
        self._count_open += 1
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        self._count_open -= 1
        if self._count_open <= 0:
            await self._async_client.__aexit__(exc_type, exc_val, exc_tb)
            self._async_client = None
            self._client_open = None

    @_throttle
    @functools.wraps(httpx.AsyncClient.get)
    async def get(self, *args, **kwargs):
        return await self._client_open.get(*args, **kwargs)

    @_throttle
    @functools.wraps(httpx.AsyncClient.put)
    async def put(self, *args, **kwargs):
        return await self._client_open.put(*args, **kwargs)

    @_throttle
    @functools.wraps(httpx.AsyncClient.post)
    async def post(self, *args, **kwargs):
        return await self._client_open.post(*args, **kwargs)

    @_throttle
    @functools.wraps(httpx.AsyncClient.request)
    async def request(self, *args, **kwargs):
        return await self._client_open.request(*args, **kwargs)

    @_throttle
    @functools.wraps(httpx.AsyncClient.head)
    async def head(self, *args, **kwargs):
        return await self._client_open.head(*args, **kwargs)

    @_throttle
    @functools.wraps(httpx.AsyncClient.options)
    async def options(self, *args, **kwargs):
        return await self._client_open.options(*args, **kwargs)

    @_throttle
    @functools.wraps(httpx.AsyncClient.delete)
    async def delete(self, *args, **kwargs):
        return await self._client_open.delete(*args, **kwargs)



import logging
from typing import Any, List, Dict
from typing import Optional

from ..core import SessionCore, URLs
from ..core import join_url
from ..core import check_session_config
from ..core import check_session_client
from ..core.constants import LOGGER_NAME
from ..core.constants import PRICE
from ..core.constants import PRODUCT
from ..core.helpers import check_response


LOGGER = logging.getLogger(LOGGER_NAME)


async def get_portfolio(session: SessionCore) -> Dict[str, Any]:
    """
    Get portfolio web call.


    Returns
    -------

    Example dictionary portfolio:

    .. code-block:: python

        {'portfolio': {'isAdded': True,
         'lastUpdated': 1088,
         'name': 'portfolio',
         'value': [
            {'id': '8614787',
             'isAdded': True,
             'name': 'positionrow',
             'value': [
                 {'isAdded': True,
                  'name': 'id',
                  'value': '8614787'},
                 {'isAdded': True,
                  'name': 'positionType',
                  'value': 'PRODUCT'},
                 {'isAdded': True,
                  'name': 'size',
                  'value': 100},
                 {'isAdded': True,
                  'name': 'price',
                  'value': 73.0},
                 {'isAdded': True,
                  'name': 'value',
                  'value': 7300.0},
                 {'isAdded': True,
                  'name': 'accruedInterest'},
                 {'isAdded': True,
                  'name': 'plBase',
                  'value': {'EUR': -6716.901595272}},
                 {'isAdded': True,
                  'name': 'todayPlBase',
                  'value': {'EUR': -7300.0}},
                 {'isAdded': True,
                  'name': 'portfolioValueCorrection',
                  'value': 0},
                 {'isAdded': True,
                  'name': 'breakEvenPrice',
                  'value': 68.15},
                 {'isAdded': True,
                  'name': 'averageFxRate',
                  'value': 1},
                 {'isAdded': True,
                  'name': 'realizedProductPl',
                  'value': 98.098404728},
                 {'isAdded': True,
                  'name': 'realizedFxPl',
                  'value': 0},
                 {'isAdded': True,
                  'name': 'todayRealizedProductPl',
                  'value': 0.0},
                 {'isAdded': True,
                  'name': 'todayRealizedFxPl',
                  'value': 0}
                  ]
            },
                    ...
        },
         {'id': 'EUR',
          'isAdded': True,
          'name': 'positionrow',
          'value': [
            {'isAdded': True,
            'name': 'id',
            'value': 'EUR'},
            {'isAdded': True,
            'name': 'positionType',
            'value': 'CASH'},
            {'isAdded': True,
            'name': 'size',
            'value': -53676.25},
            {'isAdded': True,
            'name': 'price',
            'value': 1},
            {'isAdded': True,
            'name': 'value',
            'value': -53676.25},
            {'isAdded': True,
            'name': 'accruedInterest'},
            {'isAdded': True,
            'name': 'plBase',
            'value': {'EUR': 53676.2467863145}},
            {'isAdded': True,
            'name': 'todayPlBase',
            'value': {'EUR': 53676.2467863145}},
            {'isAdded': True,
            'name': 'portfolioValueCorrection',
            'value': 0},
            {'isAdded': True,
            'name': 'breakEvenPrice',
            'value': 0},
            {'isAdded': True,
            'name': 'averageFxRate',
            'value': 1},
            {'isAdded': True,
            'name': 'realizedProductPl',
            'value': 0},
            {'isAdded': True,
            'name': 'realizedFxPl',
            'value': 0},
            {'isAdded': True,
            'name': 'todayRealizedProductPl',
            'value': 0},
            {'isAdded': True,
            'name': 'todayRealizedFxPl',
            'value': 0}]
         },
         {'id': 'USD',
          'isAdded': True,
          'name': 'positionrow',
          'value': [{'isAdded': True,
          'name': 'id',
          'value': 'USD'},
         {'isAdded': True,
          'name': 'positionType',
          'value': 'CASH'},
         {'isAdded': True,
          'name': 'size',
          'value': 0.0},
         {'isAdded': True,
         'name': 'price',
         'value': 1},
         {'isAdded': True,
         'name': 'value',
         'value': 0.0},
         {'isAdded': True,
         'name': 'accruedInterest'},
         {'isAdded': True,
         'name': 'plBase',
         'value': {'EUR': -4.216892111}},
         {'isAdded': True,
         'name': 'todayPlBase',
         'value': {'EUR': 0.0}},
         {'isAdded': True,
         'name': 'portfolioValueCorrection',
         'value': 0},
         {'isAdded': True,
         'name': 'breakEvenPrice',
         'value': 0},
         {'isAdded': True,
         'name': 'averageFxRate',
         'value': 1},
         {'isAdded': True,
         'name': 'realizedProductPl',
         'value': 0},
         {'isAdded': True,
         'name': 'realizedFxPl',
         'value': 0},
         {'isAdded': True,
         'name': 'todayRealizedProductPl',
         'value': 0},
         {'isAdded': True,
         'name': 'todayRealizedFxPl',
         'value': 0}]},
         {'id': 'PLN',
         'isAdded': True,
         'name': 'positionrow',
         'value': [{'isAdded': True,
         'name': 'id',
         'value': 'PLN'},
         {'isAdded': True,
         'name': 'positionType',
         'value': 'CASH'},
         {'isAdded': True,
         'name': 'size',
         'value': 0.0},
         {'isAdded': True,
         'name': 'price',
         'value': 1},
         {'isAdded': True,
         'name': 'value',
         'value': 0.0},
         {'isAdded': True,
         'name': 'accruedInterest'},
         {'isAdded': True,
         'name': 'plBase',
         'value': {'EUR': 1.8128205}},
         {'isAdded': True,
         'name': 'todayPlBase',
         'value': {'EUR': 0.0}},
         {'isAdded': True,
         'name': 'portfolioValueCorrection',
         'value': 0},
         {'isAdded': True,
         'name': 'breakEvenPrice',
         'value': 0},
         {'isAdded': True,
         'name': 'averageFxRate',
         'value': 1},
         {'isAdded': True,
         'name': 'realizedProductPl',
         'value': 0},
         {'isAdded': True,
         'name': 'realizedFxPl',
         'value': 0},
         {'isAdded': True,
         'name': 'todayRealizedProductPl',
         'value': 0},
         {'isAdded': True,
         'name': 'todayRealizedFxPl',
         'value': 0}]},
         {'id': 'GBP',
         'isAdded': True,
         'name': 'positionrow',
         'value': [{'isAdded': True,
         'name': 'id',
         'value': 'GBP'},
         {'isAdded': True,
         'name': 'positionType',
         'value': 'CASH'},
         {'isAdded': True,
         'name': 'size',
         'value': 0.0},
         {'isAdded': True,
         'name': 'price',
         'value': 1},
         {'isAdded': True,
         'name': 'value',
         'value': 0.0},
         {'isAdded': True,
         'name': 'accruedInterest'},
         {'isAdded': True,
         'name': 'plBase',
         'value': {'EUR': 0.0}},
         {'isAdded': True,
         'name': 'todayPlBase',
         'value': {'EUR': 0.0}},
         {'isAdded': True,
         'name': 'portfolioValueCorrection',
         'value': 0},
         {'isAdded': True,
         'name': 'breakEvenPrice',
         'value': 0},
         {'isAdded': True,
         'name': 'averageFxRate',
         'value': 1},
         {'isAdded': True,
         'name': 'realizedProductPl',
         'value': 0},
         {'isAdded': True,
         'name': 'realizedFxPl',
         'value': 0},
         {'isAdded': True,
         'name': 'todayRealizedProductPl',
         'value': 0},
         {'isAdded': True,
         'name': 'todayRealizedFxPl',
         'value': 0}]},
         {'id': 'FLATEX_EUR',
         'isAdded': True,
         'name': 'positionrow',
         'value': [{'isAdded': True,
         'name': 'id',
         'value': 'FLATEX_EUR'},
         {'isAdded': True,
         'name': 'positionType',
         'value': 'CASH'},
         {'isAdded': True,
         'name': 'size',
         'value': 0.0},
         {'isAdded': True,
         'name': 'price',
         'value': 1},
         {'isAdded': True,
         'name': 'value',
         'value': 0.0},
         {'isAdded': True,
         'name': 'accruedInterest'},
         {'isAdded': True,
         'name': 'plBase',
         'value': {'EUR': 0.0}},
         {'isAdded': True,
         'name': 'todayPlBase',
         'value': {'EUR': 0.0}},
         {'isAdded': True,
         'name': 'portfolioValueCorrection',
         'value': 0},
         {'isAdded': True,
         'name': 'breakEvenPrice',
         'value': 0},
         {'isAdded': True,
         'name': 'averageFxRate',
         'value': 1},
         {'isAdded': True,
         'name': 'realizedProductPl',
         'value': 0},
         {'isAdded': True,
         'name': 'realizedFxPl',
         'value': 0},
         {'isAdded': True,
         'name': 'todayRealizedProductPl',
         'value': 0},
         {'isAdded': True,
         'name': 'todayRealizedFxPl',
         'value': 0}]},
         {'id': 'FLATEX_USD',
         'isAdded': True,
         'name': 'positionrow',
         'value': [{'isAdded': True,
         'name': 'id',
         'value': 'FLATEX_USD'},
         {'isAdded': True,
         'name': 'positionType',
         'value': 'CASH'},
         {'isAdded': True,
         'name': 'size',
         'value': 0.0},
         {'isAdded': True,
         'name': 'price',
         'value': 1},
         {'isAdded': True,
         'name': 'value',
         'value': 0.0},
         {'isAdded': True,
         'name': 'accruedInterest'},
         {'isAdded': True,
         'name': 'plBase',
         'value': {'EUR': 0}},
         {'isAdded': True,
         'name': 'todayPlBase',
         'value': {'EUR': 0}},
         {'isAdded': True,
         'name': 'portfolioValueCorrection',
         'value': 0},
         {'isAdded': True,
         'name': 'breakEvenPrice',
         'value': 0},
         {'isAdded': True,
         'name': 'averageFxRate',
         'value': 1},
         {'isAdded': True,
         'name': 'realizedProductPl',
         'value': 0},
         {'isAdded': True,
         'name': 'realizedFxPl',
         'value': 0},
         {'isAdded': True,
         'name': 'todayRealizedProductPl',
         'value': 0},
         {'isAdded': True,
         'name': 'todayRealizedFxPl',
         'value': 0}]},
         {'id': 'FLATEX_PLN',
         'isAdded': True,
         'name': 'positionrow',
         'value': [{'isAdded': True,
         'name': 'id',
         'value': 'FLATEX_PLN'},
         {'isAdded': True,
         'name': 'positionType',
         'value': 'CASH'},
         {'isAdded': True,
         'name': 'size',
         'value': 0.0},
         {'isAdded': True,
         'name': 'price',
         'value': 1},
         {'isAdded': True,
         'name': 'value',
         'value': 0.0},
         {'isAdded': True,
         'name': 'accruedInterest'},
         {'isAdded': True,
         'name': 'plBase',
         'value': {'EUR': 0}},
         {'isAdded': True,
         'name': 'todayPlBase',
         'value': {'EUR': 0}},
         {'isAdded': True,
         'name': 'portfolioValueCorrection',
         'value': 0},
         {'isAdded': True,
         'name': 'breakEvenPrice',
         'value': 0},
         {'isAdded': True,
         'name': 'averageFxRate',
         'value': 1},
         {'isAdded': True,
         'name': 'realizedProductPl',
         'value': 0},
         {'isAdded': True,
         'name': 'realizedFxPl',
         'value': 0},
         {'isAdded': True,
         'name': 'todayRealizedProductPl',
         'value': 0},
         {'isAdded': True,
         'name': 'todayRealizedFxPl',
         'value': 0}]},
         {'id': 'FLATEX_GBP',
         'isAdded': True,
         'name': 'positionrow',
         'value': [{'isAdded': True,
         'name': 'id',
         'value': 'FLATEX_GBP'},
         {'isAdded': True,
         'name': 'positionType',
         'value': 'CASH'},
         {'isAdded': True,
         'name': 'size',
         'value': 0.0},
         {'isAdded': True,
         'name': 'price',
         'value': 1},
         {'isAdded': True,
         'name': 'value',
         'value': 0.0},
         {'isAdded': True,
         'name': 'accruedInterest'},
         {'isAdded': True,
         'name': 'plBase',
         'value': {'EUR': 0}},
         {'isAdded': True,
         'name': 'todayPlBase',
         'value': {'EUR': 0}},
         {'isAdded': True,
         'name': 'portfolioValueCorrection',
         'value': 0},
         {'isAdded': True,
         'name': 'breakEvenPrice',
         'value': 0},
         {'isAdded': True,
         'name': 'averageFxRate',
         'value': 1},
         {'isAdded': True,
         'name': 'realizedProductPl',
         'value': 0},
         {'isAdded': True,
         'name': 'realizedFxPl',
         'value': 0},
         {'isAdded': True,
         'name': 'todayRealizedProductPl',
         'value': 0},
         {'isAdded': True,
         'name': 'todayRealizedFxPl',
         'value': 0}]}]},
         'totalPortfolio': {'isAdded': True,
         'lastUpdated': 22,
         'name': 'totalPortfolio',
         'value': [{'isAdded': True,
         'name': 'degiroCash',
         'value': -53676.25},
         {'isAdded': True,
         'name': 'flatexCash',
         'value': 0.0},
         {'isAdded': True,
         'name': 'totalCash',
         'value': -53676.25},
         {'isAdded': True,
         'name': 'totalDepositWithdrawal',
         'value': 63950.27},
         {'isAdded': True,
         'name': 'todayDepositWithdrawal',
         'value': 0},
         {'isAdded': True,
         'name': 'cashFundCompensationCurrency',
         'value': 'EUR'},
         {'isAdded': True,
         'name': 'cashFundCompensation',
         'value': 0},
         {'isAdded': True,
         'name': 'cashFundCompensationWithdrawn',
         'value': 28.79},
         {'isAdded': True,
         'name': 'cashFundCompensationPending',
         'value': 0},
         {'isAdded': True,
         'name': 'todayNonProductFees',
         'value': 0},
         {'isAdded': True,
         'name': 'totalNonProductFees',
         'value': -657.242202735},
         {'isAdded': True,
         'name': 'freeSpaceNew',
         'value': {'EUR': 35136.647442}},
         {'isAdded': True,
         'name': 'reportMargin',
         'value': 35136.647442},
         {'isAdded': True,
         'name': 'reportCreationTime',
         'value': '12:48:31'},
         {'isAdded': True,
         'name': 'reportPortfValue',
         'value': 149223.516559},
         {'isAdded': True,
         'name': 'reportCashBal',
         'value': -53676.2465},
         {'isAdded': True,
         'name': 'reportNetliq',
         'value': 95547.270059},
         {'isAdded': True,
         'name': 'reportOverallMargin',
         'value': 60410.622617},
         {'isAdded': True,
         'name': 'reportTotalLongVal',
         'value': 104456.461592},
         {'isAdded': True,
         'name': 'reportDeficit',
         'value': 50780.215092},
         {'isAdded': True,
         'name': 'marginCallStatus',
         'value': 'NO_MARGIN_CALL'},
         {'isAdded': True, 'name': 'marginCallDeadline'}]}
        }
    """
    return await get_trading_update(
        session,
        params={'portfolio': 0})


async def get_portfolio_total(session: SessionCore) -> Dict[str, Any]:
    return await get_trading_update(
        session,
        params={'totalPortfolio': 0})


async def get_products_info(
        session: SessionCore,
        products_ids: List[str]) -> Dict[str, Any]:
    """
    Get Product info Web API call.
    """
    config = check_session_config(session)
    client = check_session_client(session)
    if config.product_search_url is None:
        raise AssertionError("productSearchUrl is None:"
                             " have you called get_config?")

    LOGGER.debug('get_products_info products_ids| %s', products_ids)
    url = join_url(config.product_search_url,
                   'v5/products/info')
    async with session as httpxclient:
        response = await httpxclient.post(
            url,
            cookies=session.cookies,
            params={
                'intAccount': client.int_account,
                'sessionId': config.session_id
            },
            json=products_ids
        )
        try:
            check_response(response)
        except Exception:
            LOGGER.error('get_products_info response| %s', response)
            LOGGER.error('get_products_info url| %s', url)
            LOGGER.error('get_products_info products_ids| %s', products_ids)
            LOGGER.error('get_products_info products_ids| %s', products_ids)
            raise
        resp_json = response.json()
        LOGGER.debug('get_products_info| %s', resp_json)
    return resp_json


async def get_company_profile(
        session: SessionCore,
        isin: str) -> Dict[str, Any]:
    """
    Get company profile.
    """
    # should this url be taken from config as well?

    client = check_session_client(session)
    config = check_session_config(session)

    # Look for dgtbxdsservice in network logs for financial statements etc.
    # might have intraday data as well
    url = join_url(URLs.BASE, 'dgtbxdsservice/company-profile/v2', isin)
    async with session as httpclient:
        response = await httpclient.get(
            url,
            cookies=session.cookies,
            params={
                'intAccount': client.int_account,
                'sessionId': config.session_id
            })
    check_response(response)
    LOGGER.debug(response.json())
    return response.json()


async def get_news_by_company(
        session: SessionCore,
        isin: str,
        limit: int = 10,
        languages: List[str] = ['en'],
        offset: int = 0
) -> Dict[str, Any]:
    """
    Get news for a company.
    """
    client = check_session_client(session)
    config = check_session_config(session)
    url = URLs.get_news_by_company_url(session)
    async with session as httpxclient:
        response = await httpxclient.get(
            url,
            cookies=session.cookies,
            params={
                'isin': isin,
                'limit': limit,
                'languages': languages,
                'offset': offset,
                'intAccount': client.int_account,
                'sessionId': config.session_id
            })
    check_response(response)
    resp_json = response.json()
    LOGGER.debug("get_news_by_company| %s", resp_json)
    return resp_json


async def get_price_data(*args, **kwargs):
    "DEPRECATED: Please use get_price_series instead"
    LOGGER.warn(
            "get_price_data is deprecated, please use get_price_series "
            "instead."
            )
    return await get_price_series(*args, **kwargs)


async def get_price_series(
        session: SessionCore,
        vwdId: str,
        vwdIdentifierType: str,
        resolution: PRICE.RESOLUTION = PRICE.RESOLUTION.PT1D,
        period: PRICE.PERIOD = PRICE.PERIOD.P1MONTH,
        timezone: str = 'Europe/Paris',
        culture: str = 'fr-FR',
        data_type: PRICE.TYPE = PRICE.TYPE.PRICE
) -> Dict[str, Any]:
    """
    Get price data for a company.

    Parameters
    ----------

    data_type = 'ohlc' provides access to 'Open', 'High', 'Low', 'Close' in
    that order for each period, instead of price data.

    vwdIdentifierType can be 'issueid' or 'vwdkey'

    Returns
    -------
    Example returned JSON:

    .. code-block:: json

        {
            "requestid": "1",
            "start": "2022-01-20T00:00:00",
            "end": "2022-01-20T14:12:24",
            "resolution": "PT1M",
            "series": [
                {
                    "expires": "2022-01-20T10:12:56+01:00",
                    "data": {
                        "issueId": 360114899,
                        "companyId": 1001,
                        "name": "AIRBUS",
                        "identifier": "issueid:360114899",
                        "isin": "NL0000235190",
                        "alfa": "AIR15598",
                        "market": "XPAR",
                        "currency": "EUR",
                        "type": "AAN",
                        "quality": "REALTIME",
                        "lastPrice": 113.1,
                        "lastTime": "2022-01-21T14:12:24",
                        "absDiff": -2.62,
                        "relDiff": -0.02264,
                        "highPrice": 114.46,
                        "highTime": "2022-01-21T10:31:14",
                        "lowPrice": 112.78,
                        "lowTime": "2022-01-21T13:56:36",
                        "openPrice": 114.0,
                        "openTime": "2022-01-21T09:00:19",
                        "closePrice": 114.0,
                        "closeTime": "2022-01-21T09:00:19",
                        "cumulativeVolume": 857092.0,
                        "previousClosePrice": 115.72,
                        "previousCloseTime": "2022-01-20T17:35:03",
                        "tradingStartTime": "09:00:00",
                        "tradingEndTime": "17:40:00",
                        "tradingAddedTime": "00:10:00",
                        "lowPriceP1Y": 81.84,
                        "highPriceP1Y": 121.1,
                        "windowStart": "2022-01-20T00:00:00",
                        "windowEnd": "2022-01-20T10:11:22",
                        "windowFirst": "2022-01-20T09:00:00",
                        "windowLast": "2022-01-20T10:11:00",
                        "windowHighTime": "2022-01-20T10:11:00",
                        "windowHighPrice": 114.46,
                        "windowLowTime": "2022-01-20T10:16:00",
                        "windowLowPrice": 112.78,
                        "windowOpenTime": "2022-01-20T09:00:19",
                        "windowOpenPrice": 114.0,
                        "windowPreviousCloseTime": "2022-01-19T17:35:03",
                        "windowPreviousClosePrice": 115.72,
                        "windowTrend": -0.02264
                    },
                    "id": "issueid:360114899",
                    "type": "object"
                    "times": "2022-01-20T00:00:00",
                    "expires": "2022-01-20T10:12:56+01:00",
                    "data": [
                        [
                        540,
                        114.0
                        ],
                        [
                        541,
                        114.08
                        ],
                        [
                        542,
                        113.62
                        ],
                        [
                        543,
                        113.8
                        ],
                        ...
                        [
                        552,
                        113.7
                        ]],
                "id":"price:issueid:360114899",
                "type":"time"}]
            }
    """
    # TODO: There may be an issue with the above JSON example, review.
    if vwdIdentifierType not in ('issueid', 'vwdkey'):
        raise ValueError(
                f"vwdIdentifierType must be 'issueid' or 'vwdkey', "
                f"not {vwdIdentifierType}")

    check_session_config(session)
    url = URLs.get_price_data_url(session)
    LOGGER.debug('get_price_series url| %s', url)
    params = {
        'requestid': 1,
        'resolution': str(resolution),
        'culture': culture,
        'period': str(period),
        'series': f'{data_type}:{vwdIdentifierType}:{vwdId}',
        'format': 'json',
        'userToken': session.config.client_id
    }
    LOGGER.debug('get_price_series params| %s', params)
    async with session as client:
        # 2023: Cookies are not needed for that call.
        # Since it looks like a third party, don't share session id if not
        # needed.
        response = await client.get(url,
                                    params=params)
    check_response(response)
    resp_json = response.json()
    LOGGER.debug('get_price_series response| %s', resp_json)
    return resp_json


async def get_trading_update(
        session: SessionCore,
        params: Dict[str, int]
) -> Dict[str, Any]:
    """
    Common call to target {tradingUrl}/v5/update/{intAccount}

    This is used by other calls leveraging the same endpoint in webapi.

    Known params:
    - 'portfolio': 0
        Current portfolio information.
    - 'totalPortfolio': 0
        Aggregate values about portfolio.
    - 'orders': 0
        Current orders.
    - 'historicalOrders': 0
        Closed orders.
    - 'transactions': 0
        Executed transactions.
    """
    url = URLs.get_portfolio_url(session)
    async with session as client:
        response = await client.get(url,
                                    cookies=session._cookies,
                                    params=params)

    check_response(response)
    LOGGER.debug("get_trading_update| %s", response.json())
    return response.json()


async def search_product(
        session: SessionCore,
        search_txt: Optional[str] = None,
        *,
        product_type_id: Optional[PRODUCT.TYPEID] = None,
        country_id: Optional[str] = None,
        index_id: Optional[str] = None,
        limit: int = 50,
        offset: int = 0) -> Dict[str, Any]:
    """
    Access `product_search` endpoint.

    Returns
    -------

    Example JSON response below, will be returned as a Python dict.

    .. code-block:: JSON

        {
            "offset": 0,
            "products": [
                {
                    "active": true,
                    "buyOrderTypes": [
                        "LIMIT",
                        "MARKET",
                        "STOPLOSS",
                        "STOPLIMIT"
                    ],
                    "category": "B",
                    "closePrice": 113.3,
                    "closePriceDate": "2022-02-02",
                    "contractSize": 1.0,
                    "currency": "EUR",
                    "exchangeId": "710",
                    "feedQuality": "R",
                    "feedQualitySecondary": "CX",
                    "id": "96008",
                    "isin": "NL0000235190",
                    "name": "AIRBUS",
                    "onlyEodPrices": false,
                    "orderBookDepth": 0,
                    "orderBookDepthSecondary": 0,
                    "orderTimeTypes": [
                        "DAY",
                        "GTC"
                    ],
                    "productBitTypes": [],
                    "productType": "STOCK",
                    "productTypeId": 1,
                    "qualitySwitchFree": false,
                    "qualitySwitchFreeSecondary": false,
                    "qualitySwitchable": false,
                    "qualitySwitchableSecondary": false,
                    "sellOrderTypes": [
                        "LIMIT",
                        "MARKET",
                        "STOPLOSS",
                        "STOPLIMIT"
                    ],
                    "strikePrice": -0.0001,
                    "symbol": "AIR",
                    "tradable": true,
                    "vwdId": "360114899",
                    "vwdIdSecondary": "955000256",
                    "vwdIdentifierType": "issueid",
                    "vwdIdentifierTypeSecondary": "issueid",
                    "vwdModuleId": 1,
                    "vwdModuleIdSecondary": 2
                }
            ]
        }

    """
    check_session_config(session)
    url = URLs.get_product_search_url(session, product_type_id)
    # Example query for stocks:
    # indexId=5&stockCountryId=886&requireTotal=true&offset=0&limit=100&sortColumns=name&sortTypes=asc&intAccount=123123123&sessionId=sdfasdfasdf
    params = dict(
        offset=offset,
        limit=limit,
        intAccount=session.client.int_account,
        sessionId=session.config.session_id,
        requireTotal=True
    )
    if product_type_id is not None:
        params['productTypeId'] = product_type_id
    if country_id is not None:
        params['stockCountryId'] = country_id
    if index_id is not None:
        params['indexId'] = index_id
        if search_txt is None:
            # 202307: Searching by indexId with no 'searchTxt' triggers
            # internal server error. We must provide an empty string
            search_txt = ''
    if search_txt is not None:
        params['searchText'] = search_txt
    LOGGER.debug("webapi.search_product params| %s", params)
    async with session as client:
        response = await client.get(url,
                                    cookies=session._cookies,
                                    params=params)
    check_response(response)
    LOGGER.debug("webapi.search_product response| %s", response.json())
    return response.json()


__all__ = [
    get_portfolio.__name__,
    get_portfolio_total.__name__,
    get_news_by_company.__name__,
    get_company_profile.__name__,
    get_price_series.__name__,
]



from __future__ import annotations
import logging
import dataclasses
try:
    # Ignore type here: will fail w/ Python version <3.11
    from enum import StrEnum  # type: ignore
except ImportError:
    import enum
    # Exists only starting Python 3.11
    # Reimplement what we need from it here.

    # Type warning 'already defined', this statement is executed if prior
    # import failed: no risk of redefinition.
    class StrEnum(str, enum.Enum):  # type: ignore
        def __str__(self):
            return str.__str__(self)

from typing import Union, Optional, Any

import httpx
from jsonloader import JSONclass

from .helpers import join_url
from .constants import LOGGER_NAME
from .constants import PRODUCT
from .constants import TIMEOUT
from ..core.helpers import ThrottlingClient
from .exceptions import ContextError

LOGGER = logging.getLogger(LOGGER_NAME)
LOGGER.setLevel(logging.DEBUG)


@dataclasses.dataclass
class Credentials:
    """
    Holds credentials for Web API.

    If 2FA is enabled on the account, either totp_secret or one_time_password
    must be provided for the login to be successful.

    `totp_secret` can be obtained in a 2FA editor/app (e.g. andOTP) in the
    'Secret' field.

    If use of `one_time_password` is chosen instead, be mindful that login
    must be called promptly as one_time_password expires frequently.
    """
    username: str
    password: str

    totp_secret: Union[str, None] = None
    one_time_password: Union[str, None] = None

    def __hash__(self) -> int:
        return hash(
            '|'.join(
                ':'.join((str(k), str(v)))
                for k, v in self.__dict__.items())
                )

    def __eq__(self, other: Any) -> bool:
        if not isinstance(other, Credentials):
            raise AttributeError(
                    "other is not `Credentials` instance.")
        return hash(self) == hash(other)


@JSONclass(annotations=True, annotations_type=True)
class Config:
    """
    SessionCore config, as returned by endpoint.
    """
    client_id: int
    companies_service_url: Union[str, None]
    dictionary_url: str
    i18n_url: Optional[str]
    landing_path: Optional[str]
    latest_searched_products_url: Optional[str]
    login_url: str
    mobile_landing_path: Optional[str]
    pa_url: str
    payment_service_url: Optional[str]
    product_notes_url: Optional[str]
    product_search_url: str
    product_types_url: str
    refinitiv_agenda_url: Optional[str]
    refinitiv_clips_url: Optional[str]
    refinitiv_company_profile_url: Optional[str]
    refinitiv_company_ratios_url: Optional[str]
    refinitiv_esgs_url: Optional[str]
    refinitiv_estimates_url: Optional[str]
    refinitiv_financial_statements_url: Optional[str]
    refinitiv_insider_transactions_url: Optional[str]
    refinitiv_insiders_report_url: Optional[str]
    refinitiv_investor_url: Optional[str]
    refinitiv_news_url: Optional[str]
    refinitiv_shareholders_url: Optional[str]
    refinitiv_top_news_categories_url: Optional[str]
    reporting_url: Optional[str]
    session_id: Optional[str]
    task_manager_url: Optional[str]
    trading_url: Optional[str]
    translations_url: Optional[str]
    vwd_gossips_url: Optional[str]
    vwd_news_url: Optional[str]
    vwd_quotecast_service_url: Optional[str]


@JSONclass(annotations=True, annotations_type=True)
class PAClient:
    """
    Data Structure for PA Client
    """
    address: Union[dict, None]
    # {
    #    city = Union[str, None]
    #    country = Union[str, None]
    #    street_address = Union[str, None]
    #    street_address_number = Union[str, None]
    #    zip: 83300
    # }
    bank_account: Union[dict, None]
    # {
    #    bank_account_id = Union[str, None]
    #    bic = Union[str, None]
    #    iban = Union[str, None]
    #    status: VERIFIED
    # }
    can_upgrade: Union[bool, None]
    client_role: Union[str, None]
    contract_type: Union[str, None]
    culture: Union[str, None]
    display_language: Union[str, None]
    display_name: Union[str, None]
    effective_client_role: Union[str, None]
    email: Union[str, None]
    first_contact: Union[dict, None]
    # {
    #    countryOfBirth: Union[str, None]
    #    dateOfBirth: Union[str, None]
    #    displayName: Union[str, None]
    #    firstName: Union[str, None]
    #    gender: Union[str, None]
    #    lastName: Union[str, None]
    #    nationality: Union[str, None]
    #    placeOfBirth: Union[str, None]
    # }
    flatex_bank_account: Union[dict, None]
    # {
    #    bic = Union[str, None]
    #    iban = Union[str, None]
    # }
    id: Union[str, None]
    int_account: Union[int, None]
    is_allocation_available: Union[bool, None]
    is_am_client_active: Union[bool, None]


@dataclasses.dataclass
class SessionCore:
    JSESSIONID = 'JSESSIONID'

    config: Optional[Config] = None
    client: Optional[PAClient] = None

    _max_requests_default: int = 40
    _period_seconds_default: int = 1

    # Cookies
    _cookies: Optional[httpx.Cookies] = None
    _http_client: Optional[ThrottlingClient] = None

    @property
    def cookies(self):
        return dict(self._cookies)

    def update_throttling(
            self,
            max_requests: int = 20,
            period_seconds: float = 1
            ):
        """
        Update throttling parameters. No limit if max_requests <= 0.

        Note: going over 40 requests per second seems to trigger API bans.

        Parameters
        ----------

        max_requests
            Maximum number of requests per `period_seconds` before throttling.
            If <= 0, no limit.

        period_seconds
            Period on which to count requests.
        """
        if self._http_client is None:
            self._max_requests_default = max_requests
            self._period_seconds_default = period_seconds
        else:
            self._http_client._max_requests = max_requests
            self._http_client._period_s = period_seconds

    @cookies.setter
    def cookies(self, cookies: dict):
        self._cookies = httpx.Cookies(cookies)

    def __hash__(self):
        return hash(self.__dict__.values())

    async def __aenter__(self) -> ThrottlingClient:
        if self._http_client is None:
            self._http_client = ThrottlingClient(
                    max_requests=self._max_requests_default,
                    period_seconds=self._period_seconds_default,
                    timeout=TIMEOUT
                    )
        return await self._http_client.__aenter__()

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        await self._http_client.__aexit__(exc_type, exc_val, exc_tb)


class URLs:
    BASE = 'https://trader.degiro.nl'
    LOGIN = join_url(BASE, '/login/secure/login')
    LOGIN_TOTP = join_url(BASE, '/login/secure/login/totp')
    CONFIG = join_url(BASE, '/login/secure/config')
    ACCOUNT_INFO = join_url(BASE, '/trading/secure/v5/account/info')

    @staticmethod
    def get_news_by_company_url(session: SessionCore) -> str:
        "Build news_by_company url"
        config = check_session_config(session)
        if config.refinitiv_news_url is None:
            raise ContextError(
                    "Session attribute 'refinitiv_news_url' is not set."
                    )
        return join_url(config.refinitiv_news_url, 'news-by-company')

    @staticmethod
    def get_client_info_url(session: SessionCore) -> str:
        """
        Build client info url.
        """
        config = check_session_config(session)
        if config.pa_url is None:
            raise ContextError(
                    "Session attribute 'pa_url' is not set."
                    )
        return join_url(config.pa_url, 'client')

    @staticmethod
    def get_portfolio_url(session: SessionCore) -> str:
        """
        Build portfolio url, also used for orders.
        """
        config = check_session_config(session)
        client = check_session_client(session)

        if session._cookies is None:
            raise ContextError("_cookies not set in session.")

        jsessionid = session._cookies[session.JSESSIONID]

        if config.trading_url is None:
            raise ContextError(
                    "Session attribute 'trading_url' is not set."
                    )
        url = join_url(
            config.trading_url,
            f'v5/update/{client.int_account}',
            f';jsessionid={jsessionid}')
        LOGGER.debug('get_portfolio_url| %s', url)
        return url

    @staticmethod
    def get_reporting_url(session: SessionCore) -> str:
        """
        Get reporting URL. Used for orders history and transactions.
        """
        config = check_session_config(session)
        url = config.reporting_url
        if url is None:
            raise ContextError(
                    "Session attribute 'reporting_url' is not set."
                    )
        LOGGER.debug('get_reporting_url| %s', url)
        return url

    @classmethod
    def get_orders_history_url(cls, session: SessionCore) -> str:
        """
        Get reporting URL. Used for orders history.
        """
        url = join_url(cls.get_reporting_url(session), 'v4/order-history')
        LOGGER.debug('get_orders_history_url| %s', url)
        return url

    @classmethod
    def get_transactions_url(cls, session: SessionCore) -> str:
        """
        Get reporting URL. Used for orders history.
        """
        url = join_url(cls.get_reporting_url(session), 'v4/transactions')
        LOGGER.debug('get_transactions_url| %s', url)
        return url

    @staticmethod
    def get_confirm_order_url(session: SessionCore) -> str:
        """
        Build url for confirm_order.

        This url will need to be completed by caller with `confirmation_id`
        For example:

        .. code-block:: python

            confirmation_id = '98741234kjaskdjfz098234'
            confirm_order_url_base = URLs.get_confirm_order_url(session)
            join_url(confirm_order_url_base, confirmation_id)

        """
        config = check_session_config(session)

        return join_url(
            config.trading_url,
            'v5/order'
        )

    @classmethod
    def get_orders_url(cls, session: SessionCore) -> str:
        """
        Build Get Orders url
        """
        # This is is the same endpoint and url as portfolio
        return cls.get_portfolio_url(session)

    @classmethod
    def get_check_order_url(cls, session: SessionCore) -> str:
        """
        Get check order URL.
        """
        config = check_session_config(session)

        jsessionid = check_session_cookies(session)[session.JSESSIONID]

        url = join_url(
            config.trading_url,
            f'v5/checkOrder;jsessionid={jsessionid}'
        )
        LOGGER.debug('get_check_order_url| %s', url)
        return url

    @staticmethod
    def get_price_data_url(session: SessionCore) -> str:
        check_session_config(session)
        # Check if this should be pulled from session config
        return 'https://charting.vwdservices.com/hchart/v1/deGiro/data.js'

    class PRODUCT_SEARCH_TYPE(StrEnum):
        GENERIC = 'products/lookup'
        STOCKS = 'stocks'
        ETFS = 'etfs'
        BONDS = 'bonds'
        OPTIONS = 'options'
        FUTURES = 'futures'
        FUNDS = 'funds'
        LEVERAGED_PRODUCTS = 'leverageds'
        WARRANTS = 'warrants'

    @staticmethod
    def get_product_search_url(
            session: SessionCore,
            product_type_id: Optional[PRODUCT.TYPEID] = None) -> str:
        specialization: Union[str, URLs.PRODUCT_SEARCH_TYPE] = {
            PRODUCT.TYPEID.STOCK:
                URLs.PRODUCT_SEARCH_TYPE.STOCKS,
            PRODUCT.TYPEID.ETFS:
                URLs.PRODUCT_SEARCH_TYPE.ETFS,
            PRODUCT.TYPEID.OPTIONS:
                URLs.PRODUCT_SEARCH_TYPE.OPTIONS,
            PRODUCT.TYPEID.BONDS:
                URLs.PRODUCT_SEARCH_TYPE.BONDS,
            PRODUCT.TYPEID.FUNDS:
                URLs.PRODUCT_SEARCH_TYPE.FUNDS,
            PRODUCT.TYPEID.LEVERAGE_PRODUCTS:
                URLs.PRODUCT_SEARCH_TYPE.LEVERAGED_PRODUCTS,
            PRODUCT.TYPEID.WARRANTS:
                URLs.PRODUCT_SEARCH_TYPE.WARRANTS,
            # mypy doesn't handle well the error management in dict.get:
            # here we rely on dict.get to provide default value if
            # product_type_id is not found or None.
        }.get(
            product_type_id,  # type: ignore
            URLs.PRODUCT_SEARCH_TYPE.GENERIC
        )
        config = check_session_config(session)
        url = join_url(
            config.product_search_url,
            'v5',
            specialization)
        LOGGER.debug('get_product_search_url: %s', url)
        return url

    @staticmethod
    def get_product_dictionary_url(session: SessionCore) -> str:
        config = check_session_config(session)
        url = config.dictionary_url
        LOGGER.debug('get_product_search_url: %s', url)
        return url

    @classmethod
    def get_account_info_url(cls, session: SessionCore) -> str:
        client = check_session_client(session)
        url = join_url(URLs.ACCOUNT_INFO, str(client.int_account))
        return cls._add_jsessionid(session, url)

    @classmethod
    def _add_jsessionid(cls, session: SessionCore, url: str) -> str:
        check_session_config(session)
        cookies = check_session_cookies(session)
        return url + ';jsessionid={}'.format(
            cookies[session.JSESSIONID])


def check_session_config(session: SessionCore) -> Config:
    "Raise an exception if session.config is not set"
    if session.config is None:
        raise AssertionError(
            "session.config is not set. Call get_config first.")
    return session.config


def check_session_client(session: SessionCore) -> PAClient:
    "Raise an exception if session.client is not set"
    if session.client is None:
        raise AssertionError(
            "session.client is not set. Call get_client_info first.")
    return session.client


def check_session_cookies(session: SessionCore) -> httpx.Cookies:
    "Helper to get cookies from a session."
    if session._cookies is not None:
        return session._cookies
    else:
        raise ContextError("Cookies not set in session. Abort.")



import unittest
import unittest.mock
from unittest.mock import MagicMock, AsyncMock
import asyncio
import time


import degiroasync.core
from degiroasync.core import join_url
from degiroasync.core import camelcase_to_snake
from degiroasync.core import camelcase_dict_to_snake
from degiroasync.core import set_params
from degiroasync.core.helpers import ThrottlingClient


class TestLRUCacheTimed(unittest.IsolatedAsyncioTestCase):
    async def test_cached_time_async(self):
        delay = 0.1  # in s

        @degiroasync.core.lru_cache_timed(seconds=delay*2)
        async def dummy():
            await asyncio.sleep(delay)
            return 2

        start = time.time()
        res = await dummy()
        self.assertEqual(res, 2)
        self.assertGreaterEqual(time.time() - start, delay)

        start2 = time.time()
        res = await dummy()
        self.assertEqual(res, 2)
        self.assertLessEqual(time.time() - start2, delay*0.5,
                             "Looks like result was not cached.")

        await asyncio.sleep(10*delay)

        start3 = time.time()
        res = await dummy()
        self.assertEqual(res, 2)
        self.assertGreaterEqual(time.time() - start3, delay,
                                "Looks like old cached result was not removed."
                                )

    async def test_cached_time_sync(self):
        delay = 0.1  # in s

        @degiroasync.core.lru_cache_timed(seconds=delay*1.5)
        def dummy():
            time.sleep(delay)
            return 1

        start = time.time()
        dummy()
        self.assertGreaterEqual(time.time() - start, delay)

        start2 = time.time()
        dummy()
        self.assertLessEqual(time.time() - start2, delay * 0.5,
                             "Looks like result was not cached.")

        await asyncio.sleep(delay)
        start3 = time.time()
        dummy()
        self.assertGreaterEqual(time.time() - start3, delay,
                                "Looks like old cached result was not removed."
                                )


class TestDegiroAsyncJoinUrl(unittest.TestCase):
    def test_join_url(self):
        url = join_url('https://foo.bar', '/rest/of/url')
        self.assertEqual(url, 'https://foo.bar/rest/of/url')
        url = join_url('https://foo.bar/product', '/rest/of/url')
        self.assertEqual(url, 'https://foo.bar/product/rest/of/url')
        url = join_url('https://foo.bar/product/', '/rest/of/url')
        self.assertEqual(url, 'https://foo.bar/product/rest/of/url')

    def test_set_params(self):
        class Foo:
            pass

        foo = Foo()

        set_params(foo, [
                 {'isAdded': True,
                  'name': 'id',
                  'value': '8614787'},
                 {'isAdded': True,
                  'name': 'positionType',
                  'value': 'PRODUCT'},
                 {'isAdded': True,
                  'name': 'size',
                  'value': 100},
                 {'isAdded': True,
                  'name': 'price',
                  'value': 73.0},
                 {'isAdded': True,
                  'name': 'value',
                  'value': 7300.0},
            ])

        self.assertEqual(foo.id, '8614787')
        self.assertEqual(foo.positionType, 'PRODUCT')
        self.assertEqual(foo.size, 100)
        self.assertEqual(foo.price, 73.0)
        self.assertEqual(foo.value, 7300.0)

    def test_camelcase_to_snake(self):
        inp = 'iAmCamelCase'
        out = camelcase_to_snake(inp)
        self.assertEqual(out, 'i_am_camel_case')

    def test_camelcase_dict_to_snake(self):
        d = {'fooBar': 2, 'camelCase': {'camelCase': 1}}
        out = camelcase_dict_to_snake(d, recursive=False)
        self.assertEqual(out, {'foo_bar': 2, 'camel_case': {'camelCase': 1}})

    def test_camelcase_dict_to_snake_recursive(self):
        d = {'fooBar': 2, 'camelCase': {'camelCase': 1}}
        out = camelcase_dict_to_snake(d, recursive=True)
        self.assertEqual(out, {'foo_bar': 2, 'camel_case': {'camel_case': 1}})


class TestThrottlingClient(unittest.IsolatedAsyncioTestCase):
    async def test_throttling(self):
        max_requests = 2
        period_seconds = 0.2
        client = ThrottlingClient(
                max_requests=max_requests,
                period_seconds=period_seconds)
        client._async_client = MagicMock()
        client._client_open = MagicMock()
        client._client_open.get = AsyncMock()

        n_calls = 20
        calls = []
        start = time.time()
        for _ in range(n_calls):
            calls.append(client.get())
        for c in calls:
            await c
        del calls
        self.assertGreaterEqual(
                time.time() - start,
                (n_calls-max_requests) * period_seconds / max_requests)




from typing import Any, Union, List, Dict, Tuple, Optional
import datetime
import logging
import asyncio
import itertools

from jsonloader import JSONclass

from .product import ProductBase
from .product import ProductFactory
from .. import webapi
from ..core import SessionCore
from ..core import ORDER
from ..core import TRANSACTION
from ..core import LOGGER_NAME
from ..core import camelcase_dict_to_snake


LOGGER = logging.getLogger(LOGGER_NAME)


@JSONclass(annotations=True, annotations_type=True)
class Order:
    order_id: Optional[str] = None  # Can be None if STATUS == REJECTED
    product_id: str
    size: Union[float, int]
    price: float
    buysell: ORDER.ACTION
    order_type_id: int
    order_time_type_id: int

    # product: Union[ProductBase, None] = None  # do we want to reinstantiate
    # products here or let user?
    #: Creation date of the order, not always available for "Outstanding".
    created: Optional[str] = None

    current_traded_size: int = 0
    total_traded_size: int = 0
    # 'CREATE' or ... ?
    type: Optional[str] = None
    is_active: bool

    #: Not always available for "Outstanding" orders
    status: Optional[ORDER.STATUS] = None

    # {
    # 'id': 182722888,
    # 'productId': 65153,
    # 'date': '2020-02-07T09:00:10+01:00',
    # 'buysell': 'B',
    # 'price': 36.07,
    # 'quantity': 20,
    # 'total': -721.4,
    # 'orderTypeId': 0,
    # 'counterParty': 'MK',
    # 'transfered': False,
    # 'fxRate': 0,
    # 'totalInBaseCurrency': -721.4,
    # 'feeInBaseCurrency': -0.29,
    # 'totalPlusFeeInBaseCurrency': -721.69,
    # 'transactionTypeId': 0,
    # 'tradingVenue': 'XPAR'
    # }


@JSONclass(annotations=True, annotations_type=True)
class Transaction:
    id: str
    product: ProductBase
    date: datetime.datetime
    buysell: ORDER.ACTION
    price: float
    quantity: float
    total: float
    transfered: bool
    fx_rate: float
    total_in_base_currency: float
    total_plus_fee_in_base_currency: float
    total_plus_all_fees_in_base_currency: float
    transaction_type_id: Union[TRANSACTION.TYPEID, int]

# Raw JSON Example
# {"id":335133382,"productId":8615503,"date":"2022-02-26T00:00:00+01:00","buysell":"S","price":134.6600,"quantity":-40,"total":5386.4000000000,"transfered":true,"fxRate":0,"nettFxRate":0,"grossFxRate":0,"autoFxFeeInBaseCurrency":0,"totalInBaseCurrency":5386.4000000000,"totalFeesInBaseCurrency":0,"totalPlusFeeInBaseCurrency":5386.4000000000,"totalPlusAllFeesInBaseCurrency":5386.4000000000,"transactionTypeId":20


async def confirm_order():
    """
    Not implemented. Use webapi.confirm_order.
    """
    raise NotImplementedError


async def check_order(
        session: SessionCore,
        *,
        product: ProductBase,
        buy_sell: ORDER.ACTION,
        time_type: ORDER.TIME,
        order_type: ORDER.TYPE,
        size: int,
        price: Union[float, None] = None,
        ) -> Dict[str, Any]:
    """
    This must be called to obtain a confirmation_id prior to confirming an
    order.

    This can also be used to get an order fees before confirming the order.

    .. code-block:: python

        # Get your products through search_product
        check_order(
            product=product,
            buy_sell=ORDER.ACTION.SELL,
            time_type=ORDER.TIME,
            order_type=ORDER.TYPE,
            size=1,
            price=100
        )

    WARNING: This call is rate limited at the end-point level, tests would show
    the call to be rate limited at 1 per second. Users should throttle their
    calls to this function.

    Example return:

    .. code-block:: python

        {
            'confirmation_id': '0f404158-3628-414b-87fc-91e2ab2ba1ee',
            'free_space_new': 62283.1,
            'transaction_fee': 0.5,
            'show_ex_ante_report_link': True
        }

    In the case of a SELL order, the parameter 'transaction_opposite_fee'
    has also been observed. It does not seem to serve a differente purpose
    than transaction_fee.
    """
    assert isinstance(buy_sell, ORDER.ACTION)

    resp_json = await webapi.check_order(
        session=session,
        product_id=product.base.id,
        buy_sell=buy_sell,
        time_type=time_type,
        order_type=order_type,
        size=size,
        price=price
    )
    return camelcase_dict_to_snake(resp_json['data'])


async def get_orders(
        session: SessionCore,
        from_date: Union[datetime.datetime, None] = None,
        to_date: Union[datetime.datetime, None] = None,
        ) -> Tuple[List[Order]]:
    """
    Get current orders and history.

    to_date
        Request orders history up to `to_date`. Defaults to today.
    from_date
        Request orders history from `from_date`. Defaults to today - 7 days.

    Return current_orders, historical_orders.

    """
    if to_date is None:
        to_date = datetime.datetime.today()
    if from_date is None:
        from_date = to_date - datetime.timedelta(days=7)

    orders_current_resp, orders_history_resp = await asyncio.gather(
        webapi.get_orders(session),
        webapi.get_orders_history(
            session,
            from_date=from_date.strftime(webapi.ORDER_DATE_FORMAT),
            to_date=to_date.strftime(webapi.ORDER_DATE_FORMAT))
    )

    LOGGER.debug(
            "get_orders orders_current_resp| %s",
            orders_current_resp)
    orders_dict = orders_current_resp['orders']
    del orders_current_resp
    for order in orders_dict:
        if 'isActive' not in order and 'created' not in order:
            # If no information and not created, consider order not active.
            # It was chosen to mark isActive = False explicitly instead of
            # having an Optional is_active attribute as a None might easily
            # be misinterpreted as False by clients.
            order['isActive'] = False

    orders_history_dict = orders_history_resp['data']
    LOGGER.debug("get_orders orders_dict| %s", orders_dict)
    LOGGER.debug("get_orders orders_history_dict| %s", orders_history_dict)
    # Ensure types expected types & set constants.
    for order in itertools.chain(orders_dict, orders_history_dict):
        order['productId'] = str(order['productId'])
        order['buysell'] = {
            'B': ORDER.ACTION.BUY,
            'S': ORDER.ACTION.SELL
        }[order['buysell']]
        if 'status' in order:
            order['status'] = {
                'CONFIRMED': ORDER.STATUS.CONFIRMED,
                'REJECTED': ORDER.STATUS.REJECTED,
                'PENDING': ORDER.STATUS.PENDING
                    }[order['status']]
    # 2022.03 mypy does not handle __new__ well, ignore for now.
    return (
        [Order(camelcase_dict_to_snake(o))  # type: ignore
            for o in orders_dict],
        [Order(camelcase_dict_to_snake(o))  # type: ignore
            for o in orders_history_dict]
    )


async def get_transactions(
        session: SessionCore,
        from_date: Union[datetime.datetime, None] = None,
        to_date: Union[datetime.datetime, None] = None
        ) -> List[Transaction]:
    """
    Get transactions for `session`.

    from_date
        Request transactions from `from_date`. Defaults to `to_date - 7 days`.

    to_date
        Request transactions to `to_date`. Defaults to today.
    """
    to_date = to_date or datetime.datetime.now()
    from_date = from_date or to_date - datetime.timedelta(days=7)

    resp = await webapi.get_transactions(
        session,
        from_date=from_date.strftime(webapi.ORDER_DATE_FORMAT),
        to_date=to_date.strftime(webapi.ORDER_DATE_FORMAT)
    )
    data = resp['data'].copy()
    products_gen = ProductFactory.init_batch(
        session,
        map(lambda t: {'id': str(t['productId'])}, data))
    products = [p async for p in products_gen]
    del products_gen

    async def _build_transaction(prod, trans):
        trans.update(dict(
            id=str(trans['id']),
            product=prod,
            date=datetime.datetime.fromisoformat(trans['date']),
            buysell={'B': ORDER.ACTION.BUY,
                     'S': ORDER.ACTION.SELL}[trans['buysell']],
        ))
        return Transaction(camelcase_dict_to_snake(trans))

    transactions = await asyncio.gather(*[_build_transaction(p, t)
                                          for p, t in zip(products, data)])
    return transactions
